#!/usr/bin/env python
"""
_Run_

Unittest for the WMCore.DataStructs.Run class

"""

import unittest

from WMCore.DataStructs.Run import Run


class RunTest(unittest.TestCase):
    """
    _RunTest_

    """

    def testA(self):
        """instantiation"""

        run1 = Run()
        self.assertEqual(run1.run, None)
        self.assertEqual(len(run1), 0)

        run2 = Run(1000000)
        self.assertEqual(run2.run, 1000000)
        self.assertEqual(len(run2), 0)

        run3 = Run(1000000, 1)
        self.assertEqual(run3.run, 1000000)
        self.assertEqual(len(run3), 1)
        self.assertEqual(run3[0], 1)

        run4 = Run(1000000, 1, 2, 3, 4, 5)
        self.assertEqual(run4.run, 1000000)
        self.assertEqual(len(run4), 5)
        self.assertEqual(run4[0], 1)
        self.assertEqual(run4[4], 5)

        # using new DBS lumi/event format
        run5 = Run(555, [1, 2, 3])
        self.assertEqual(run5.run, 555)
        self.assertEqual(len(run5), 3)
        self.assertDictEqual(run5.eventsPerLumi, {1: None, 2: None, 3: None})

        run6 = Run(666, [(1, 11), (2, 22), (3, 33)])
        self.assertEqual(run6.run, 666)
        self.assertEqual(len(run6), 3)
        self.assertDictEqual(run6.eventsPerLumi, {1: 11, 2: 22, 3: 33})

        run7 = Run(666, [(1, 11), (2, 22), (3, 33), (1, 4), (3, 2)])
        self.assertEqual(run7.run, 666)
        self.assertEqual(len(run7), 3)
        self.assertDictEqual(run7.eventsPerLumi, {1: 15, 2: 22, 3: 35})

        run8 = Run(666, [(1, None), (2, None), (3, None)])
        self.assertEqual(run8.run, 666)
        self.assertEqual(len(run8), 3)
        self.assertDictEqual(run8.eventsPerLumi, {1: None, 2: None, 3: None})

        run10 = Run(555, *[1, 2, 3])
        self.assertEqual(run10.run, 555)
        self.assertEqual(len(run10), 3)
        self.assertDictEqual(run10.eventsPerLumi, {1: None, 2: None, 3: None})

        run11 = Run(666, *[(1, None), (2, None), (3, None)])
        self.assertEqual(run11.run, 666)
        self.assertDictEqual(run11.eventsPerLumi, {1: None, 2: None, 3: None})
        self.assertEqual(len(run11), 3)

        run13 = Run(666, *[(1, 11), (2, 22), (3, 33), (1, 4), (3, 2)])
        self.assertEqual(run13.run, 666)
        self.assertEqual(len(run13), 3)
        self.assertDictEqual(run13.eventsPerLumi, {1: 15, 2: 22, 3: 35})

    def test2(self):
        """
        test equality & addition operators

        """
        run1 = Run(1, 1, 2, 3, 4, 5)
        run1copy = Run(1, 1, 2, 3, 4, 5)
        run1difflumi = Run(1, 6, 7, 8, 9, 10)
        run1overlap = Run(1, 3, 4, 5, 6, 7)
        run2 = Run(2, 1, 2, 3, 4, 5)

        self.assertTrue(run1 == run1copy)
        self.assertFalse(run1 == run1difflumi)
        self.assertFalse(run1 == run1overlap)
        self.assertFalse(run1 == run2)
        self.assertTrue(run1 != run2)

        run1 += run1difflumi
        self.assertEqual(len(run1), 10)
        for l in run1difflumi:
            self.assertTrue(l in run1)

    def testC(self):
        """
        comparision and sorting

        """
        def cmp_runs(run_x, run_y):
            return [
                run_x < run_y, run_x <= run_y,
                run_x == run_y, run_x != run_y,
                run_x > run_y, run_x >= run_y,]

        run1 = Run(1, 1, 2, 3)
        run2 = Run(2, 4, 5, 6)
        run3 = Run(3, 7, 8, 9)

        runlist = list(sorted([run2, run3, run1]))
        self.assertListEqual(runlist, [run1, run2, run3])
        self.assertListEqual(cmp_runs(run1, run2),
                             [True, True, False, True, False, False])

        run4 = Run(4, 1, 2, 3)
        run5 = Run(4, 4, 5, 6)
        run6 = Run(4, 7, 8, 9)

        runlist = list(sorted([run5, run6, run4]))
        self.assertListEqual(runlist, [run4, run5, run6])

        runlist = list(sorted([run5, run1, run3, run4, run6, run2]))
        self.assertListEqual(runlist, [run1, run2, run3, run4, run5, run6])

        run7  = Run(9, 1, 2)
        run8  = Run(9, 1, 2, 3)
        run9  = Run(9, 1, 2, 3)
        run10 = Run(9, 1, 2, 3, 4)
        run11 = Run(9, 1, 2, 4)
        run12 = Run(9, 1, 4)
        run13 = Run(9, 2)
        run14 = Run(9, 3, 4, 5)
        run15 = Run(9, 8, 9, 10)

        runlist = list(sorted([run10, run15, run8,  run11, run13,
                               run7,  run9,  run14, run12]))
        self.assertListEqual(runlist,
                            [run7, run8, run9, run10, run11,
                             run12, run13, run14, run15])

        self.assertListEqual(cmp_runs(run14, run15),
                             [True, True, False, True, False, False])
        self.assertListEqual(cmp_runs(run8, run8),
                             [False, True, True, False, False, True])
        self.assertListEqual(cmp_runs(run8, run9),
                             [False, True, True, False, False, True])

        run20 = Run(666, [(1, 11), (2, 22), (3, 33)])
        run21 = Run(666, [(1, 11), (2, 22), (3, 33)])
        run22 = Run(666, [(1, 11), (2, 22), (3, 32)])
        run23 = Run(666, [(1, 11), (2, 22), (3, 34)])
        run24 = Run(666, [(1, 11), (2, 21), (3, 33)])
        run25 = Run(666, [(1, 11), (2, 21), (3, 34)])
        run26 = Run(666, [(1, 11), (2, 23), (3, 32)])

        self.assertListEqual(cmp_runs(run20, run21),
                             [False, True, True, False, False, True])
        self.assertListEqual(cmp_runs(run20, run22),
                             [False, False, False, True, True, True])
        self.assertListEqual(cmp_runs(run20, run23),
                             [True, True, False, True, False, False])
        self.assertListEqual(cmp_runs(run20, run24),
                             [False, False, False, True, True, True])
        self.assertListEqual(cmp_runs(run20, run25),
                             [False, False, False, True, True, True])
        self.assertListEqual(cmp_runs(run20, run26),
                             [True, True, False, True, False, False])

    def testD(self):
        """
        test hashing for sets

        """
        run1 = Run(1, 1, 2, 3)
        run2 = Run(2, 4, 5, 6)
        run3 = Run(3, 7, 8, 9)
        run4 = Run(4, 1, 2, 3)
        run5 = Run(4, 4, 5, 6)
        run6 = Run(4, 7, 8, 9)
        run7 = Run(9, 1, 2, 3)
        run8 = Run(9, 1, 2, 3)
        run9 = Run(9, 2)
        run10 = Run(9, 3, 4, 5)
        run11 = Run(9, 8, 9, 10)

        s = set()
        s.add(run1)
        s.add(run2)
        s.add(run3)
        s.add(run4)
        s.add(run5)
        s.add(run6)
        s.add(run7)
        s.add(run8)
        s.add(run9)
        s.add(run10)
        s.add(run11)
        self.assertEqual(run7, run8)
        self.assertNotEqual(run7, run9)
        self.assertNotEqual(run7, run10)

    def testNewHash(self):
        run1 = Run(1, 1, 2, 3)
        self.assertEqual(hash(run1), 79038151938585768)
        run666 = Run(666, [(1, 111), (3, 33), (2, 22)])
        self.assertEqual(hash(run666), 21721583371131476)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_GetFinishedWorkflows_

Oracle implementation of Workflow.GetFinishedWorkflows

Created on Aug 30, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Workflow.GetFinishedWorkflows import GetFinishedWorkflows as MySQLGetFinishedWorkflows

class GetFinishedWorkflows(MySQLGetFinishedWorkflows):
    pass



#!/usr/bin/env python
"""
_SiteLocalConfig_t_

Unit test for parsing the site local config file.
"""
from __future__ import print_function

import os
import unittest

from nose.plugins.attrib import attr

from WMCore.Storage.SiteLocalConfig import SiteLocalConfig, SiteConfigError
from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig
from WMCore.WMBase import getTestBase


class SiteLocalConfigTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testFNALSiteLocalConfig(self):
        """
        _testFNALSiteLocalConfig_

        Verify that the FNAL site config file is parsed correctly.
        """
        os.environ['SITECONFIG_PATH'] = '/cvmfs/cms.cern.ch/SITECONF/T1_US_FNAL'
        fnalConfigFileName = os.path.join(getTestBase(),
                                          "WMCore_t/Storage_t",
                                          "T1_US_FNAL_SiteLocalConfig.xml")
        mySiteConfig = SiteLocalConfig(fnalConfigFileName)

        assert mySiteConfig.siteName == "T1_US_FNAL", "Error: Wrong site name."
        assert len(list(mySiteConfig.eventData)) == 1, "Error: Wrong number of event data keys."
        assert mySiteConfig.eventData[
                   "catalog"] == "trivialcatalog_file:/cvmfs/cms.cern.ch/SITECONF/T1_US_FNAL_Disk/PhEDEx/storage.xml?protocol=fallbackxrd", \
            "Eroror: Event data catalog is wrong."

        goldenServers = ["http://cmsfrontier.cern.ch:8000/FrontierInt",
                         "http://cmsfrontier.cern.ch:8000/FrontierInt",
                         "http://cmsfrontier1.cern.ch:8000/FrontierInt",
                         "http://cmsfrontier2.cern.ch:8000/FrontierInt",
                         "http://cmsfrontier3.cern.ch:8000/FrontierInt",
                         "http://cmsfrontier4.cern.ch:8000/FrontierInt"]
        for frontierServer in mySiteConfig.frontierServers:
            assert frontierServer in goldenServers, \
                "Error: Unknown server: %s" % frontierServer
            goldenServers.remove(frontierServer)

        assert len(goldenServers) == 0, \
            "Error: Missing frontier servers."

        goldenProxies = ["http://cmsfrontier1.fnal.gov:3128",
                         "http://cmsfrontier2.fnal.gov:3128",
                         "http://cmsfrontier3.fnal.gov:3128",
                         "http://cmsfrontier4.fnal.gov:3128"]
        for frontierProxy in mySiteConfig.frontierProxies:
            assert frontierProxy in goldenProxies, \
                "Error: Unknown proxy: %s" % frontierProxy
            goldenProxies.remove(frontierProxy)

        assert len(goldenProxies) == 0, \
            "Error: Missing proxy servers."

        assert mySiteConfig.stageOuts[0]["command"] == "xrdcp", \
            "Error: Wrong stage out command."
        assert mySiteConfig.stageOuts[0]["protocol"] == "XRootD", \
            "Error: Protocol is not correct."
        assert mySiteConfig.stageOuts[0]["option"] == "-p", \
            "Error: option is not correct."
        # assert False
        return

    def testLoadingConfigFromOverridenEnvVarriable(self):
        """
        test SiteLocalConfig module method loadSiteLocalConfig when loading
        site config from location defined by WMAGENT_SITE_CONFIG_OVERRIDE
        env. variable

        """
        vandyConfigFileName = os.path.join(getTestBase(),
                                           "WMCore_t/Storage_t",
                                           "T1_US_FNAL_SiteLocalConfig.xml")
        os.environ["WMAGENT_SITE_CONFIG_OVERRIDE"] = vandyConfigFileName
        os.environ["SITECONFIG_PATH"] = "/cvmfs/cms.cern.ch/SITECONF/T1_US_FNAL"

        # still using legacy trivial catalog
        mySiteConfig = loadSiteLocalConfig()
        self.assertEqual(mySiteConfig.siteName, "T1_US_FNAL",
                         "Error: Wrong site name.")

    # this test requires access to CVMFS
    @attr("integration")
    def testSlcPhedexNodesEqualPhedexApiNodes(self):
        """
        For each site, verify that the stageout node specified in
        site-local-config.xml is the same as the one returned by the PhEDEx api.
        """
        os.environ["CMS_PATH"] = "/cvmfs/cms.cern.ch"

        nodes = ['FIXME']

        for d in os.listdir("/cvmfs/cms.cern.ch/SITECONF/"):
            # Only T0_, T1_... folders are needed
            if d[0] == "T":
                os.environ["SITECONFIG_PATH"] = "/cvmfs/cms.cern.ch/SITECONF/%s" % (d)
                os.environ[
                    'WMAGENT_SITE_CONFIG_OVERRIDE'] = '/cvmfs/cms.cern.ch/SITECONF/%s/JobConfig/site-local-config.xml' % (
                    d)
                try:
                    # still using legacy trivial catalog
                    slc = loadSiteLocalConfig()
                except SiteConfigError as e:
                    print(e.args[0])
                phedexNode = slc.localStageOut.get("phedex-node")
                self.assertTrue(phedexNode in nodes,
                                "Error: Node specified in SLC (%s) not in list returned by PhEDEx api" % phedexNode)
        return


if __name__ == "__main__":
    unittest.main()



from Integration_t.RequestLifeCycleBase_t import RequestLifeCycleBase_t
import unittest

class ReRecoLifeCycle_t(unittest.TestCase, RequestLifeCycleBase_t):

    requestParams = {'CMSSWVersion' : 'CMSSW_4_4_2_patch2',
                        "Scenario": "pp", "inputMode" : "Scenario", "ProcScenario" : "pp",
                        "RequestType": "ReReco",
                        "Requestor": 'integration',
                        "InputDataset": "/BTag/Run2011B-v1/RAW", "RunWhitelist" : [177316, 177317],
                        "GlobalTag": 'FT_R_44_V11::All', 'Group' : 'DMWM', "RequestPriority" : 10,
                    }


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
File       : DBSConcurrency.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: dedicated module to holds DBS related functions executed
concurrent calls to DBS APIs.
"""

import json
import urllib
from WMCore.Services.pycurl_manager import getdata as multi_getdata


def getBlockInfo4PU(blockNames, dbsUrl, ckey, cert):
    """
    Fetch block information details, file list and number of events, from DBS
    server. Here we use concrete set of parameters for DBS to use in this case, i.e.
    we must look-up only valid files and get full details from the DBS API (in order
    to get number of events).
    :param blockNames: list of block names
    :param dbsUrl: dbs URL
    :param ckey: user keyfile
    :param cert: user certificate
    :return: dictionary of {block: {"FileList": list of strings, "NumberOfEvents": integer}, ...}
    """
    urls = []
    for blk in blockNames:
        # need to encode block name properly
        block = urllib.parse.quote_plus(blk)
        url = f"{dbsUrl}/files?detail=true&validFileOnly=1&block_name={block}"
        urls.append(url)
    # place concurrent calls to DBS, please note that multi_getdata is generator, therefore
    # it does not put DBS results into the memory until this generator is iterated
    results = multi_getdata(urls, ckey, cert)
    # parse output of getdata in some form
    blockInfo = {}
    for row in results:
        blk = row['url'].split('block_name=')[-1]
        block = urllib.parse.unquote_plus(blk)
        data = json.loads(row['data'])
        files = [r['logical_file_name'] for r in data]
        nevents = sum([r['event_count'] for r in data])
        blockInfo[block] = {'FileList': files, 'NumberOfEvents': nevents}
    return blockInfo



#!/usr/bin/env python
"""
_DBSUtils_t_

Unit test for the DBSUtils module
"""

import time
import unittest

from nose.plugins.attrib import attr

from WMCore.Services.DBS.DBSUtils import urlParams
from WMCore.Services.DBS.DBS3Reader import DBS3Reader


class DBSUtilsTest(unittest.TestCase):
    """
    DBSUtilsTest represent unit test class
    """

    def testUrlParams(self):
        """
        urlParams should return dictionary of URL parameters
        """
        url = 'http://a.b.com?d=1&f=bla'
        results = urlParams(url)
        self.assertCountEqual(results, {'d': '1', 'f': 'bla'})
        self.assertTrue(results.get('d'), 1)
        self.assertTrue(results.get('f'), 'bla')

        url = 'http://a.b.com?d=1&f=bla&d=2'
        results = urlParams(url)
        self.assertCountEqual(results, {'d': ['1', '2'], 'f': 'bla'})

    @attr("integration")
    def testGetParallelListDatasetFileDetails(self):
        """
        test parallel execution of listDatasetFileDetails DBS API
        We use small dataset with the following characteristics:

                dasgoclient -query="dataset=/VBF1Parked/HIRun2013A-v1/RAW summary" | jq
                [
                  {
                        "file_size": 6053097,
                        "nblocks": 7,
                        "nevents": 0,
                        "nfiles": 7,
                        "nlumis": 428,
                        "num_block": 7,
                        "num_event": 0,
                        "num_file": 7,
                        "num_lumi": 428
                  }
                ]

                The parallel call should perform better then sequential while results
        should remain the same.

        """
        url = "https://cmsweb.cern.ch/dbs/prod/global/DBSReader"
        reader1 = DBS3Reader(url, logger=None, parallel=False, aggregate=True)
        reader2 = DBS3Reader(url, logger=None, parallel=True, aggregate=True)
        dataset = '/VBF1Parked/HIRun2013A-v1/RAW'
        time0 = time.time()
        res1 = reader1.listDatasetFileDetails(dataset)
        time1 = time.time() - time0
        self.assertTrue(time1 > 0)  # to avoid pyling complaining about not used varaiable
        time0 = time.time()
        res2 = reader2.listDatasetFileDetails(dataset)
        time2 = time.time() - time0
        self.assertTrue(time2 > 0)  # to avoid pyling complaining about not used varaiable
        self.assertTrue(res1 == res2)

    @attr("integration")
    def testGetParallelListFileBlockLocation(self):
        """
        test parallel execution of listFileBlockLocation DBS API
        We use small dataset with the following data:

        dasgoclient -query="block dataset=/VBF1Parked/HIRun2013A-v1/RAW"

        The parallel call should perform better then sequential while results
        should remain the same.

        """
        url = "https://cmsweb.cern.ch/dbs/prod/global/DBSReader"
        reader1 = DBS3Reader(url, logger=None, parallel=False, aggregate=True)
        reader2 = DBS3Reader(url, logger=None, parallel=True, aggregate=True)
        blocks = [
            '/VBF1Parked/HIRun2013A-v1/RAW#52da10be-5c87-11e2-912f-842b2b4671d8',
            '/VBF1Parked/HIRun2013A-v1/RAW#6861d50a-5c7f-11e2-912f-842b2b4671d8',
            '/VBF1Parked/HIRun2013A-v1/RAW#6dd88910-5c80-11e2-912f-842b2b4671d8',
            '/VBF1Parked/HIRun2013A-v1/RAW#6e258f12-5c80-11e2-912f-842b2b4671d8',
            '/VBF1Parked/HIRun2013A-v1/RAW#fc64292c-5c81-11e2-912f-842b2b4671d8',
            '/VBF1Parked/HIRun2013A-v1/RAW#fc87c8dc-5c81-11e2-912f-842b2b4671d8',
            '/VBF1Parked/HIRun2013A-v1/RAW#fcd9876c-5c81-11e2-912f-842b2b4671d8'
            ]
        time0 = time.time()
        res1 = reader1.listFileBlockLocation(blocks)
        time1 = time.time() - time0
        self.assertTrue(time1 > 0)  # to avoid pyling complaining about not used varaiable
        time0 = time.time()
        res2 = reader2.listFileBlockLocation(blocks)
        time2 = time.time() - time0
        self.assertTrue(time2 > 0)  # to avoid pyling complaining about not used varaiable
        self.assertTrue(res1 == res2)

    @attr("integration")
    def testGetParallelGetParentFilesGivenParentDataset(self):
        """
        test parallel execution of getParentFilesGivenParentDataset DBS API
        We use small the following data:

        # find lfn for some dataset
        dasgoclient -query="file dataset=/VBF1Parked/Run2012D-22Jan2013-v1/AOD"
        ...
        /store/data/Run2012D/VBF1Parked/AOD/22Jan2013-v1/120000/F64DFA15-15A8-E211-9277-80000048FE80.root

        # find parent dataset
        dasgoclient -query="parent dataset=/VBF1Parked/Run2012D-22Jan2013-v1/AOD"
        /VBF1Parked/Run2012D-v1/RAW

        The parallel call should perform better then sequential while results
        should remain the same.

        """
        url = "https://cmsweb.cern.ch/dbs/prod/global/DBSReader"
        reader1 = DBS3Reader(url, logger=None, parallel=False, aggregate=True)
        reader2 = DBS3Reader(url, logger=None, parallel=True, aggregate=True)
        parentDataset = '/VBF1Parked/Run2012D-v1/RAW'
        childLFN = '/store/data/Run2012D/VBF1Parked/AOD/22Jan2013-v1/120000/F64DFA15-15A8-E211-9277-80000048FE80.root'
        time0 = time.time()
        res1 = reader1.getParentFilesGivenParentDataset(parentDataset, childLFN)
        time1 = time.time() - time0
        self.assertTrue(time1 > 0)  # to avoid pyling complaining about not used varaiable
        time0 = time.time()
        res2 = reader2.getParentFilesGivenParentDataset(parentDataset, childLFN)
        time2 = time.time() - time0
        self.assertTrue(time2 > 0)  # to avoid pyling complaining about not used varaiable
        self.assertTrue(res1 == res2)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""
scramulator.py

Scram emulator for unittests of the CMSSW runtime environments

Created by Dave Evans on 2010-03-15.
Copyright (c) 2010 Fermilab. All rights reserved.
"""
from __future__ import print_function

import sys
import os


def dispatcher(*args):
    """
    _dispatcher_

    Examine argument list and route to the appropriate call

    """
    if args[0] == '--arch':
        args = args[2:]

    if args[0] == 'project':
        scramProject(*args)
    elif args[0] == 'runtime':
        scramRuntime(*args)
    elif args[0] == 'ru':
        scramRuntime(*args)
    else:
        print("Unknown scram command: %s\nFull command: %s" % (args[0], args))
        sys.exit(1)


def scramProject(*args):
    print("Emulating scram project command...")
    workingDir = os.getcwd()
    projectVersion = args[2]
    newDir = os.path.join(os.getcwd(), projectVersion)
    if not os.path.exists(newDir):
        os.makedirs(newDir)

    sys.exit(0)


def scramRuntime(*args):
    print("export GREETING=\"Hello World\";")


if __name__ == '__main__':
    dispatcher(*sys.argv[1:])



#!/usr/bin/env python
"""
_ListCurrentSites_

Query the database to get all the current sites in Resource Control
"""

from WMCore.Database.DBFormatter import DBFormatter

class ListCurrentSites(DBFormatter):
    sql = """SELECT site_name AS site
                    FROM wmbs_location
           """

    def execute(self, conn = None, transaction = False):
        results = self.dbi.processData(self.sql,
                                       conn = conn, transaction = transaction)
        return self.formatDict(results)



#!/usr/bin/env python
"""
_GetWorkflowTask_

MySQL implementation of Jobs.GetWorkflowTask
"""

from WMCore.Database.DBFormatter import DBFormatter


class GetWorkflowTask(DBFormatter):
    sql = ('SELECT '
           ' wmbs_workflow.name, wmbs_workflow.task, wmbs_workflow.id as taskid, wmbs_job.id, '
           ' wmbs_workflow.type, wmbs_sub_types.name AS subtype FROM wmbs_workflow '
           'INNER JOIN wmbs_subscription ON wmbs_workflow.id = wmbs_subscription.workflow '
           'INNER JOIN wmbs_jobgroup ON wmbs_subscription.id = wmbs_jobgroup.subscription '
           'INNER JOIN wmbs_job ON wmbs_jobgroup.id = wmbs_job.jobgroup '
           'INNER JOIN wmbs_sub_types ON wmbs_subscription.subtype = wmbs_sub_types.id '
           'WHERE wmbs_job.id = :jobid')

    def execute(self, jobIDs, conn=None, transaction=False):
        binds = []
        for jobID in jobIDs:
            binds.append({"jobid": jobID})

        result = self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)
        return self.formatDict(result)



#!/usr/bin/env python

"""
CMSSWFetcher

A unittest for seeing if we can pull code the configCache to the WN
Written by someone who has no idea what CMSSWFetcher is supposed to do.
"""
import os
import os.path
import unittest

from WMCore.WMSpec import WMTask
from WMCore.WMSpec import WMStep
from WMCore.WMSpec.Steps.Fetchers.CMSSWFetcher import CMSSWFetcher
from WMQuality.TestInitCouchApp                import TestInitCouchApp as TestInit
from WMCore.Cache.WMConfigCache                import ConfigCache

class CMSSWFetcherTest(unittest.TestCase):
    """
    Main test for the URLFetcher

    """

    def setUp(self):
        """
        Basic setUp

        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setupCouch("config_test", "GroupUser", "ConfigCache")

        self.testDir = self.testInit.generateWorkDir()

        return

    def tearDown(self):
        """
        Basic tearDown

        """
        self.testInit.delWorkDir()
        self.testInit.tearDownCouch()
        return

    def getConfig(self):
        """
        _getConfig_

        Create a test config and put it in the cache
        """
        PSetTweak = {'someKey': "Hello, I am a PSetTweak.  It's nice to meet you."}

        configCache = ConfigCache(os.environ["COUCHURL"], couchDBName = 'config_test')
        configCache.createUserGroup(groupname = "testGroup", username = 'testOps')
        configCache.setPSetTweaks(PSetTweak = PSetTweak)
        configCache.attachments['configFile'] = 'This Is A Test Config'
        configCache.save()
        return configCache


    def createTask(self, configCache):
        """
        _createTask_

        Create a test task that includes the
        fileURL
        """

        task = WMTask.makeWMTask("testTask")
        task.makeStep("step1")
        task.makeStep("step2")

        for t in task.steps().nodeIterator():
            t = WMStep.WMStepHelper(t)
            os.mkdir(os.path.join(self.testDir, t.name()))
            t.setStepType("CMSSW")
            t.data.application.section_('command')
            t.data.application.configuration.configCacheUrl = configCache.dburl
            t.data.application.configuration.cacheName      = configCache.dbname
            t.data.application.configuration.configId       = configCache.getCouchID()
            t.data.application.command.psetTweak            = 'tweak'
            t.data.application.command.configuration        = 'configCache.file'
        return task

    def testA_BasicFunction(self):
        """
        _BasicFunction_

        Run a test to find out if we can grab a configCache
        """
        configCache = self.getConfig()
        task        = self.createTask(configCache = configCache)
        fetcher     = CMSSWFetcher()
        fetcher.setWorkingDirectory(workingDir = self.testDir)
        self.assertEqual(fetcher.workingDirectory(), self.testDir)

        fetcher(wmTask = task)
        configFilePath = os.path.join(self.testDir, 'step2', 'configCache.file')
        self.assertTrue(os.path.isfile(configFilePath))

        with open(configFilePath) as f:
            content = f.read()

        self.assertEqual(content, 'This Is A Test Config')
        return

if __name__ == "__main__":
    unittest.main()



from __future__ import (print_function, division)
import os
import unittest

from mock import mock

from WMCore.Storage.Backends.LCGImpl import LCGImpl


class LCGImplTest(unittest.TestCase):
    def setUp(self):
        self.LCGImpl = LCGImpl()
        self.LCGImpl.setups = self.setups = "test setups"
        self.timeoutOptions = self.LCGImpl.timeoutOptions = \
            '--srm-timeout 600 --sendreceive-timeout 600 --connect-timeout 300'

    @mock.patch('WMCore.Storage.Backends.LCGImpl.os')
    def testInit(self, mock_os):
        mock_os.environ.get.side_effect = lambda x: True if x == "GLITE_LOCATION" or x == "GRID_ENV_LOCATION" else False
        mock_os.path.isfile.return_value = True
        mock_os.path.normpath.side_effect = lambda x: "home" + x
        mock_os.path.join.side_effect = lambda a, b: os.path.join("", b)
        testLCGImpl = LCGImpl(True)
        self.assertTrue(testLCGImpl.stageIn)
        setups = []
        setups += ['source home/../etc/profile.d/grid-env.sh; ']
        setups += ['source home/grid-env.sh; ']
        setups += ['date "+%Y-%m-%dT%H:%M:%S"; ']
        self.assertTrue(testLCGImpl.stageIn)
        for setup in setups:
            self.assertIn(setup, testLCGImpl.setups)

    def testCreateSourceName_startswithSlash(self):
        self.assertEqual("file://name", self.LCGImpl.createSourceName("protocol", "//name"))

    @mock.patch('WMCore.Storage.Backends.LCGImpl.os.path')
    def testCreateSourceName_isfile(self, mock_path):
        mock_path.isfile.return_value = True
        mock_path.abspath.return_value = "/some/path"
        self.assertEqual("file:/some/path", self.LCGImpl.createSourceName("protocol", "name"))

    def testCreateSourceName_simple(self):
        self.assertEqual("name", self.LCGImpl.createSourceName("protocol", "name"))

    def testCreateStageOutCommand_stageInFile(self):
        self.LCGImpl.stageIn = True
        result = self.LCGImpl.createStageOutCommand("srm://sourcePFN", "file://targetPFN")
        expectedResults = self.getStageOutCommandResults("srm://sourcePFN", "file://targetPFN", "srm://sourcePFN",
                                                         "//targetPFN", "/bin/rm -f //targetPFN", False)
        self.assertEqual(expectedResults, result)

    def testCreateStageOutCommand_stageISrmCvmfs(self):
        self.LCGImpl.stageIn = True
        result = self.LCGImpl.createStageOutCommand("srm://sourcePFN", "srm://targetPFN", options="cvmfs")
        createRemoveFileCommandResult = "%s lcg-del -b -l -D srmv2 %s --vo cms srm://targetPFN" % \
                                        (self.setups, self.timeoutOptions)
        expectedResults = self.getStageOutCommandResults("srm://sourcePFN", "srm://targetPFN", "srm://sourcePFN",
                                                         "srm://targetPFN", createRemoveFileCommandResult, True)
        self.assertEqual(expectedResults, result)

    @mock.patch('WMCore.Storage.Backends.LCGImpl.StageOutImpl.createRemoveFileCommand')
    def testCreateStageOutCommand_options(self, mock_createRemoveFileCommand):
        mock_createRemoveFileCommand.return_value = "command"
        result = self.LCGImpl.createStageOutCommand("file://sourcePFN", "targetPFN", options="test")
        expectedResults = self.getStageOutCommandResults("file://sourcePFN", "targetPFN",
                                                         "targetPFN", "//sourcePFN", "command", False, options="test")
        self.assertEqual(expectedResults, result)

    @mock.patch('WMCore.Storage.Backends.LCGImpl.StageOutImpl.createRemoveFileCommand')
    def testCreateStageOutCommand_checksum(self, mock_createRemoveFileCommand):
        mock_createRemoveFileCommand.return_value = "command"
        result = self.LCGImpl.createStageOutCommand("file://sourcePFN", "targetPFN", checksums={"adler32": "32"})
        expectedResults = self.getStageOutCommandResults("file://sourcePFN", "targetPFN",
                                                         "targetPFN", "//sourcePFN", "command", False,
                                                         checksums="00000032")
        self.assertEqual(expectedResults, result)

    def getStageOutCommandResults(self, sourcePFN, targetPFN, remotePFN, localPFN,
                                  createRemoveFileCommandResult, useCVMFS, options=None, checksums=None):
        result = "#!/bin/sh\n"

        copyCommand = "lcg-cp -b -D srmv2 --vo cms --srm-timeout 2400 --sendreceive-timeout 2400 --connect-timeout 300 --verbose"
        if options != None:
            copyCommand += " %s " % options
        copyCommand += " %s " % sourcePFN
        copyCommand += " %s 2> stageout.log" % targetPFN
        if useCVMFS:
            result += "(\n"
            result += "echo Modifying PATH and LD_LIBRARY_PATH to remove /cvmfs/cms.cern.ch elements\n"
            result += "export PATH=`echo $PATH | sed -e 's+:*/cvmfs/cms.cern.ch/[^:]*++'g`\n"
            result += "export LD_LIBRARY_PATH=`echo $LD_LIBRARY_PATH | sed -e 's+:*/cvmfs/cms.cern.ch/[^:]*++'g`\n"
            result += "echo Sourcing CVMFS UI setup script\n"
            result += ". /cvmfs/grid.cern.ch/emi3ui-latest/etc/profile.d/setup-ui-example.sh\n"

            result += copyCommand
        else:
            result += self.setups
            result += copyCommand

        result += """
            EXIT_STATUS=$?
            cat stageout.log
            echo -e "\nlcg-cp exit status: $EXIT_STATUS"
            if [[ $EXIT_STATUS != 0 ]]; then
                echo "ERROR: lcg-cp exited with $EXIT_STATUS"
                echo "Cleaning up failed file:"
                %s
                exit $EXIT_STATUS
            fi

            """ % createRemoveFileCommandResult

        result += "FILE_SIZE=`stat -c %s"
        result += " %s`\n" % localPFN
        result += "echo \"Local File Size is:  $FILE_SIZE\"\n"

        if checksums:
            checksumCommand = \
                """
                if [[ "X$SRM_CHECKSUM" != "X" ]]; then
                    if [[ "$SRM_CHECKSUM" == "%s" ]]; then
                        exit 0
                    else
                        echo "ERROR: Checksum Mismatch between local and SE"
                        echo "Cleaning up failed file"
                        %s
                        exit 60311
                    fi
                fi
                exit 0
                """ % (checksums, createRemoveFileCommandResult)
        else:
            checksumCommand = "exit 0"

        metadataCheck = \
            """
            LCG_OUTPUT=`lcg-ls -l -b -D srmv2 %s %s 2>/dev/null`
            SRM_SIZE=`echo "$LCG_OUTPUT" | awk 'NR==1{print $5}'`
            SRM_CHECKSUM=`echo "$LCG_OUTPUT" | sed -nr 's/^.*\s([a-f0-9]{8})\s*\([aA][dD][lL][eE][rR]32\)\s*$/\\1/p'`
            echo "Remote File Size is: $SRM_SIZE"
            echo "Remote Checksum is:  $SRM_CHECKSUM"
            if [[ $SRM_SIZE == $FILE_SIZE ]]; then
                %s
            else
                echo $LCG_OUTPUT
                echo "ERROR: Size Mismatch between local and SE. Cleaning up failed file..."
                %s
                exit 60311
            fi
            """ % (self.timeoutOptions, remotePFN, checksumCommand, createRemoveFileCommandResult)
        result += metadataCheck
        if useCVMFS:
            result += ")\n"
        return result

    @mock.patch('WMCore.Storage.StageOutImpl.StageOutImpl.executeCommand')
    def testRemoveFile(self, mock_executeCommand):
        self.LCGImpl.removeFile("file")
        mock_executeCommand.assert_called_with("%s lcg-del -b -l -D srmv2 %s --vo cms file" %
                                               (self.setups, self.timeoutOptions))



#!/bin/env python





from __future__ import print_function
from builtins import str

import unittest
import time

from WMCore.Services.UUIDLib import makeUUID

class UUIDTest(unittest.TestCase):


    def setUp(self):
        pass

    def tearDown(self):
        pass


    def testUUID(self):

        listOfIDs = []

        tmpID = makeUUID()
        splitID = None
        splitID = tmpID.split('-')

        for i in range(0,1000):
            tmpID = makeUUID()
            tmpSplit = tmpID.split('-')
            self.assertNotEqual(tmpSplit[1], splitID[1], "Second component of UUID the same %s != %s"
                             % (tmpSplit[1], splitID[1]))
            self.assertNotEqual(tmpSplit[4], splitID[4], "Fourth component of UUID the same %s != %s"
                             % (tmpSplit[4], splitID[4]))
            self.assertEqual(type(tmpID), str)
            self.assertEqual(listOfIDs.count(tmpID), 0, "UUID repeated!  %s found in list %i times!"
                             % (tmpID, listOfIDs.count(tmpID)))
            listOfIDs.append(tmpID)



        return


    def testTime(self):

        nUIDs     = 100000
        startTime = time.time()
        for i in range(0,nUIDs):
            makeUUID()
        print("We can make %i UUIDs in %f seconds" %(nUIDs, time.time() - startTime))

if __name__ == '__main__':
    unittest.main()



"""
File       : MSPileupTasks_t.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Unit tests for MicorService/MSPileup/MSPileupTasks.py module
"""

# system modules
import os
import time
import logging
import unittest

# rucio modules
from rucio.client import Client

# WMCore modules
from WMQuality.Emulators.RucioClient.MockRucioApi import MockRucioApi
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMCore.MicroService.MSPileup.MSPileupTasks import MSPileupTasks
from WMCore.MicroService.MSPileup.MSPileupData import MSPileupData
from WMCore.MicroService.MSPileup.MSPileupMonitoring import MSPileupMonitoring
from WMCore.MicroService.Tools.Common import getMSLogger
from WMCore.Services.Rucio.Rucio import Rucio
from Utils.Timers import gmtimeSeconds


class TestRucioClient(Client):
    """Fake implementation for Rucio client"""
    def __init__(self, account='test', logger=None, state='OK'):
        self.account = 'test'
        if logger:
            self.logger = logger
        else:
            self.logger = getMSLogger(False)
        self.rses = ['rse1', 'rse2', 'T2_XX_CERN']
        self.state = state
        self.doc = {'id': '123', 'rse_expression': 'T2_XX_CERN', 'state': self.state}

    def list_replication_rules(self, kwargs):
        """Immitate list_replication_rules Rucio client API"""
        # we will mock rucio doc based on provided state
        # the states are: OK, REPLICATING, STUCK, SUSPENDED, WAITING_APPROVAL, INJECT
        doc = dict(self.doc)
        if self.state != 'OK':
            doc['locks_ok_cnt'] = 2
            doc['locks_replicating_cnt'] = 2
            doc['locks_stuck_cnt'] = 2
        for key, val in kwargs.items():
            doc[key] = val
        docs = [doc]
        self.logger.info("### TestRucioClient docs %s", docs)
        if 'Exception' in self.state:
            raise Exception(self.state)
        return docs

    def delete_replication_rule(self, rid):
        """Immidate delete replication_rule Rucio client API"""
        msg = f"delete rule ID {rid}"
        self.logger.info(msg)

    def add_replication_rule(self, dids, copies, rses, **kwargs):
        """Immitate add replication rule Rucio client API"""
        msg = f"add replication rule for {dids}, copies {copies} rses {rses}"
        self.logger.info(msg)

    def list_rses(self, rseExpression):
        """Immitate get list_rses Rucio client API"""
        for rse in self.rses:
            yield {'rse': rse}

    def get_rse_usage(self, rse):
        """Immitate get rse usage Rucio client API"""
        # it is a generator which provides information about given RSE
        doc = {'id': '123', 'source': 'unavailable',
               'used': 440245583794751,
               'free': None,
               'total': 440245583794751, 'files': 138519,
               'rse': rse}
        yield doc

    def listDataRules(self, pname, **kwargs):
        """Mock the Rucio wrapper listDataRules API"""
        self.logger.info("%s: mocking listDataRules.", self.__class__.__name__)
        return self.list_replication_rules(kwargs)

    def evaluateRSEExpression(self, expr):
        """Mock the Rucio wrapper evaluateRSEExpression API"""
        self.logger.info("%s: mocking evaluateRSEExpression.", self.__class__.__name__)
        return self.rses


class MSPileupTasksTest(EmulatedUnitTestCase):
    """Unit test for MSPileupTasks module"""

    def setUp(self):
        """
        setup Unit tests
        """
        # we will define log stream to capture everything that goes to the log stream
        self.logger = logging.getLogger()
        self.userDN = 'test-dn'

        # setup rucio client
        self.rucioAccount = 'wmcore_pileup'
        self.hostUrl = 'http://cms-rucio-int.cern.ch'
        self.authUrl = 'https://cms-rucio-auth-int.cern.ch'
        creds = {"client_cert": os.getenv("X509_USER_CERT", "Unknown"),
                 "client_key": os.getenv("X509_USER_KEY", "Unknown")}
        configDict = {'rucio_host': self.hostUrl, 'auth_host': self.authUrl,
                      'creds': creds, 'auth_type': 'x509'}

        # setup rucio wrapper
        testRucioClient = TestRucioClient(logger=self.logger, state='OK')
        self.rucioClient = Rucio(self.rucioAccount, configDict=configDict, client=testRucioClient)

        self.validRSEs = ['rse1', 'rse2']

        # setup pileup data manager
        msConfig = {'reqmgr2Url': 'http://localhost',
                    'rucioAccount': 'wmcore_pileup',
                    'rucioUrl': 'http://cms-rucio-int.cern.ch',
                    'rucioAuthUrl': 'https://cms-rucio-auth-int.cern.ch',
                    'mongoDB': 'msPileupDB',
                    'mongoDBCollection': 'msPileupDBCollection',
                    'mongoDBServer': 'mongodb://localhost',
                    'mongoDBReplicaSet': '',
                    'mongoDBUser': None,
                    'mongoDBPassword': None,
                    'mockMongoDB': True}
        self.mgr = MSPileupData(msConfig, skipRucio=True)
        self.monMgr = MSPileupMonitoring(msConfig)

        # setup our pileup data
        self.pname = '/primary/processed/PREMIX'
        pname = self.pname
        fullReplicas = 3
        campaigns = ['c1', 'c2']
        data = {
            'pileupName': pname,
            'pileupType': 'classic',
            'expectedRSEs': self.validRSEs,
            'currentRSEs': self.validRSEs,
            'fullReplicas': fullReplicas,
            'campaigns': campaigns,
            'containerFraction': 1.0,
            'replicationGrouping': "ALL",
            'active': True,
            'pileupSize': 0,
            'ruleIds': ['rse1']}
        self.data = data

        self.mgr.createPileup(data, self.validRSEs, userDN=self.userDN)

        # add more docs similar in nature but with different size
        data['pileupName'] = pname.replace('processed', 'processed-2')
        self.mgr.createPileup(data, self.validRSEs, userDN=self.userDN)
        data['pileupName'] = pname.replace('processed', 'processed-3')
        self.mgr.createPileup(data, self.validRSEs, userDN=self.userDN)

    def testMSPileupTasks(self):
        """
        Unit test for MSPileupTasks
        """
        self.logger.info("---------- CHECK for state=OK -----------")

        obj = MSPileupTasks(self.mgr, self.monMgr, self.logger, self.rucioAccount, self.rucioClient)
        obj.monitoringTask()

        # we added three pileup documents and should have update at least one of them
        # in our report, so we check for update pileup message in report
        report = obj.getReport()
        found = False
        for doc in report.getDocuments():
            if 'update pileup' in doc['entry']:
                found = True
        self.assertEqual(found, True)

        # at this step the T2_XX_CERN should be added to currentRSEs as it is provided
        # by TestRucioClient class via list_replication_rules
        spec = {'pileupName': self.pname}
        results = self.mgr.getPileup(spec)
        self.assertEqual(len(results), 1)
        doc = results[0]
        self.assertEqual('T2_XX_CERN' in doc['currentRSEs'], True)
        obj.activeTask()
        obj.inactiveTask()

        # get report documents and log them accordingly
        report = obj.getReport()
        for uuid, entries in report.getReportByUuid().items():
            msg = f"-------- task {uuid} --------"
            self.logger.info(msg)
            for item in entries:
                self.logger.info(item)

        # now we can test non OK state in Rucio
        self.logger.info("---------- CHECK for state=STUCK -----------")
        self.rucioClient = TestRucioClient(logger=self.logger, state='STUCK')
        obj = MSPileupTasks(self.mgr, self.monMgr, self.logger, self.rucioAccount, self.rucioClient)
        obj.monitoringTask()
        # at this step the T2_XX_CERN should NOT be added to currentRSEs
        spec = {'pileupName': self.pname}
        results = self.mgr.getPileup(spec)
        self.assertEqual(len(results), 1)
        doc = results[0]
        self.assertEqual('T2_XX_CERN' in doc['currentRSEs'], False)
        obj.activeTask()
        obj.inactiveTask()

        # now we can test how our code will behave with rucio exceptions
        self.logger.info("---------- CHECK for state=CustomException -----------")

        # we use CustomException for state to check how our code will
        # handle Rucio API exceptions
        self.rucioClient = TestRucioClient(logger=self.logger, state='CustomException')
        obj = MSPileupTasks(self.mgr, self.monMgr, self.logger, self.rucioAccount, self.rucioClient)
        obj.monitoringTask()

    def testMSPileupTasksWithMockApi(self):
        """
        Unit test for MSPileupTasks with RucioMockApi
        """
        # we may take some mock data from
        # https://github.com/dmwm/WMCore/blob/master/test/data/Mock/RucioMockData.json
        # e.g. /MinimumBias/ComissioningHI-v1/RAW' dataset
        pname = '/MinimumBias/ComissioningHI-v1/RAW'
        data = dict(self.data)
        data['pileupName'] = pname

        # add transition record as it now requires for update API
        trec = {'DN': 'localhost-test', 'containerFraction': 1, 'customDID': 'customDID', 'updateTime': gmtimeSeconds()}
        data.update({'transition': [trec]})

        self.mgr.createPileup(data, self.validRSEs, userDN=self.userDN)

        # now create mock rucio client
        rucioClient = MockRucioApi(self.rucioAccount, hostUrl=self.hostUrl, authUrl=self.authUrl)
        obj = MSPileupTasks(self.mgr, self.monMgr, self.logger, self.rucioAccount, rucioClient)
        obj.monitoringTask()
        obj.activeTask()
        obj.inactiveTask()

        # we added new pileup document and should have update pileup message in report
        report = obj.getReport()
        found = False
        for doc in report.getDocuments():
            if 'update pileup' in doc['entry']:
                found = True
        self.assertEqual(found, True)

        # update doc in MSPileup and call cleanup task to delete it
        data['active'] = False
        data['rulesIds'] = []
        data['currentRSEs'] = []
        data['deactivatedOn'] = 0
        self.mgr.updatePileup(data)
        obj.cleanupTask(0)

    def testMSPileupTasksWithCustomName(self):
        """
        Unit test for MSPileupTasks with RucioMockApi and customName DID
        """
        self.logger.info("---------- testMSPileupTasksWithCustomName ----------")
        # we may take some mock data from
        # https://github.com/dmwm/WMCore/blob/master/test/data/Mock/RucioMockData.json
        # e.g. /Cosmics/ComissioningHI-PromptReco-v1/RECO dataset
        pname = '/Cosmics/ComissioningHI-PromptReco-v1/RECO'
        data = dict(self.data)
        data['pileupName'] = self.pname + 'CUSTOM'
        data['customName'] = pname

        # add transition record as it now requires for update API
        trec = {'DN': 'localhost-test', 'containerFraction': 1, 'customDID': 'customDID', 'updateTime': gmtimeSeconds()}
        data.update({'transition': [trec]})

        self.mgr.createPileup(data, self.validRSEs, userDN=self.userDN)

        # now create mock rucio client
        rucioClient = MockRucioApi(self.rucioAccount, hostUrl=self.hostUrl, authUrl=self.authUrl)
        obj = MSPileupTasks(self.mgr, self.monMgr, self.logger, self.rucioAccount, rucioClient)
        obj.monitoringTask()
        obj.activeTask()
        obj.inactiveTask()
        cmsDict, cusDict = obj.pileupSizeTask()

        # we should have out pname in both cmsDict and cusDict maps
        # the asserts are commented as we do not have customName records in mongoDB yet
        # self.assertEqual(len(cusDict), 1)
        # self.assertTrue(len(cmsDict) > 1)
        # instead we'll only print their numbers such that we can verify it during integration tests
        self.logger.info("pileupSizeTask has cmsDict=%s, cusDict=%s records", len(cmsDict), len(cusDict))

        # we added new pileup document and should have update pileup message in report
        report = obj.getReport()
        found = False
        for doc in report.getDocuments():
            self.logger.info("### doc %s", doc)
            if 'update pileup' in doc['entry']:
                found = True
        self.assertEqual(found, True)

    def testPartialPileupTask(self):
        """
        Unit test for partialPileupTask with RucioMockApi and customName DID
        """
        self.logger.info("---------- testPartialPileupTask ----------")

        # clean up MongoDB for this test
        for doc in self.mgr.getPileup({}):
            self.mgr.deletePileup(doc)

        # create mock rucio client
        rucioClient = MockRucioApi(self.rucioAccount, hostUrl=self.hostUrl, authUrl=self.authUrl)

        # create new pileup document
        # step 1 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        pname = '/primary/processed-test/PREMIX'
        self.logger.info("### step 1: create partial pileup document %s", pname)
        data = dict(self.data)
        data['pileupName'] = pname
        self.mgr.createPileup(data, self.validRSEs, userDN=self.userDN)

        # sleep a little bit between creation and updated document
        time.sleep(1)

        # at this point we should have a document with transition record
        # step 2 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        spec = {'pileupName': pname}
        res = self.mgr.getPileup(spec)
        self.logger.info("### step 2: getPileup %s", res)
        self.assertEqual(len(res), 1)
        record = res[0]
        self.assertEqual(record['customName'], '')
        self.assertEqual(record['containerFraction'], 1.0)
        self.assertEqual(len(record['transition']), 1)
        trRec = record['transition'][0]
        self.assertEqual(trRec['customDID'], pname)
        self.assertEqual(trRec['containerFraction'], 1.0)

        # step 3 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        obj = MSPileupTasks(self.mgr, self.monMgr, self.logger, self.rucioAccount, rucioClient)
        obj.partialPileupTask()

        # at this point nothing should changed with pileup document
        res = self.mgr.getPileup(spec)
        self.logger.info("### step 3: getPileup %s", res)
        self.assertEqual(len(res), 1)
        record = res[0]
        self.assertEqual(record['customName'], '')
        self.assertEqual(record['containerFraction'], 1.0)
        self.assertEqual(len(record['transition']), 1)
        trRec = record['transition'][0]
        self.assertEqual(trRec['customDID'], pname)
        self.assertEqual(trRec['containerFraction'], 1.0)

        # step 4 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        # add new spec with transition change
        spec = {'pileupName': pname, 'containerFraction': 0.5}
        res = self.mgr.updatePileup(spec, rseList=self.validRSEs)
        self.logger.info("### step 4: updatePileup %s", res)
        self.assertEqual(len(res), 0)

        # step 5 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        # after this step the last transition record should be updated

        # at this point nothing should changed with pileup document
        res = self.mgr.getPileup(spec)
        self.logger.info("### step 5: getPileup %s", res)
        self.assertEqual(len(res), 1)
        record = res[0]

        self.assertEqual(record['customName'], '')
        self.assertEqual(record['containerFraction'], 0.5)
        self.assertEqual(len(record['transition']), 2)
        trRec = record['transition'][0]
        self.assertEqual(trRec['customDID'], pname)
        self.assertEqual(trRec['containerFraction'], 1.0)
        trRec = record['transition'][1]
        self.assertEqual(trRec['customDID'], pname + '-V1')
        self.assertEqual(trRec['containerFraction'], 1.0)

        # step 6 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        # it should fetch new rules and update document accordingly since fraction has changed
        obj.partialPileupTask()

        # step 7 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        # we should have updated customName, and last transition record fraction
        spec = {'pileupName': pname}
        res = self.mgr.getPileup(spec)
        self.logger.info("### step 7: getPileup %s", res)
        self.assertEqual(len(res), 1)
        record = res[0]
        self.assertEqual(record['customName'], pname + '-V1')
        self.assertEqual(record['containerFraction'], 0.5)
        self.assertEqual(len(record['transition']), 2)
        trRec = record['transition'][0]
        self.assertEqual(trRec['customDID'], pname)
        self.assertEqual(trRec['containerFraction'], 1.0)
        trRec = record['transition'][1]
        self.assertEqual(trRec['customDID'], pname + '-V1')
        self.assertEqual(trRec['containerFraction'], 0.5)

        # step 8 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        spec = {'pileupName': pname, 'containerFraction': 0.75}
        res = self.mgr.updatePileup(spec, rseList=self.validRSEs)
        self.logger.info("### step 8: updatePileup %s", res)
        self.assertEqual(len(res), 0)

        # step 9 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        # at this point we should have 3 transition records
        res = self.mgr.getPileup(spec)
        self.logger.info("### step 9: getPileup %s", res)
        self.assertEqual(len(res), 1)
        record = res[0]
        self.assertEqual(record['customName'], pname + '-V1')
        self.assertEqual(record['containerFraction'], 0.75)
        self.assertEqual(len(record['transition']), 3)
        trRec = record['transition'][0]
        self.assertEqual(trRec['customDID'], pname)
        self.assertEqual(trRec['containerFraction'], 1.0)
        trRec = record['transition'][1]
        self.assertEqual(trRec['customDID'], pname + '-V1')
        self.assertEqual(trRec['containerFraction'], 0.5)
        trRec = record['transition'][2]
        self.assertEqual(trRec['customDID'], pname + '-V2')
        self.assertEqual(trRec['containerFraction'], 0.5)

        # step 10 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        # it should fetch new rules and update document accordingly since fraction has changed
        obj.partialPileupTask()

        # step 11 of https://gist.github.com/amaltaro/b4f9bafc0b58c10092a0735c635538b5
        # here we should have new custom name, and updated contrainer fraction of last record
        res = self.mgr.getPileup(spec)
        self.logger.info("### step 11: getPileup %s", res)
        self.assertEqual(len(res), 1)
        record = res[0]
        self.assertEqual(record['customName'], pname + '-V2')
        self.assertEqual(record['containerFraction'], 0.75)
        self.assertEqual(len(record['transition']), 3)
        trRec = record['transition'][0]
        self.assertEqual(trRec['customDID'], pname)
        self.assertEqual(trRec['containerFraction'], 1.0)
        trRec = record['transition'][1]
        self.assertEqual(trRec['customDID'], pname + '-V1')
        self.assertEqual(trRec['containerFraction'], 0.5)
        trRec = record['transition'][2]
        self.assertEqual(trRec['customDID'], pname + '-V2')
        self.assertEqual(trRec['containerFraction'], 0.75)

        # last check for transition records
        for idx, rec in enumerate(record['transition']):
            if idx == 0:
                self.assertEqual(rec['customDID'], pname)
                self.assertEqual(rec['containerFraction'], 1.0)
            elif idx == 1:
                self.assertEqual(rec['customDID'], pname + '-V1')
                self.assertEqual(rec['containerFraction'], 0.5)
            elif idx == 2:
                self.assertEqual(rec['customDID'], pname + '-V2')
                self.assertEqual(rec['containerFraction'], 0.75)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
_SetPendingJobSlotsForSite_

Oracle implementation of ResourceControl.SetPendingJobSlotsForSite

Created on Fri Jun 22 14:16:13 2012

@author: dballest
"""

from WMCore.ResourceControl.MySQL.SetPendingJobSlotsForSite import \
            SetPendingJobSlotsForSite as MySQLSetPendingJobSlotsForSite

class SetPendingJobSlotsForSite(MySQLSetPendingJobSlotsForSite):
    pass



#!/usr/bin/env python

from __future__ import print_function

import json
import os
import unittest
from http.client import HTTPException

from WMCore_t.ReqMgr_t.TestConfig import config

from WMCore.REST.Test import fake_authz_headers
from WMQuality.REST.RESTBaseUnitTestWithDBBackend import RESTBaseUnitTestWithDBBackend

ADMIN = {'role': ["admin_role"], 'group': ["admin_group"]}
OPS = {'role': ["ops_role"], 'group': ["ops_group"]}
PPD = {'role': ["ppd_role"], 'group': ["ppd_group"]}


# this needs to move in better location
def insertDataToCouch(couchUrl, couchDBName, data):
    import WMCore.Database.CMSCouch
    server = WMCore.Database.CMSCouch.CouchServer(couchUrl)
    database = server.connectDatabase(couchDBName)

    doc = database.commit(data)
    return doc


def getAuthHeader(hmacData, reqAuth):
    roles = {}
    for role in reqAuth['role']:
        roles[role] = {'group': reqAuth['group']}

    return fake_authz_headers(hmacData, dn="/TEST/DN/CN", roles=roles, format="dict")


class ReqMgrTest(RESTBaseUnitTestWithDBBackend):
    """
    Test WorkQueue Service client
    It will start WorkQueue RESTService
    Server DB sets from environment variable.
    Client DB sets from environment variable.

    This checks whether DS call makes without error and return the results.
    Not the correctness of functions. That will be tested in different module.

    """

    def setFakeDN(self):
        # put into ReqMgr auxiliary database under "software" document scram/cmsms
        # which we'll need a little for request injection
        # Warning: this assumes the same structure in jenkins wmcore_root/test
        self.admin_header = getAuthHeader(self.test_authz_key.data, ADMIN)
        self.ops_header = getAuthHeader(self.test_authz_key.data, OPS)
        self.ppd_header = getAuthHeader(self.test_authz_key.data, PPD)

    def setUp(self, initRoot=True):
        self.setConfig(config)
        self.setCouchDBs([(config.views.data.couch_reqmgr_db, "ReqMgr"),
                          (config.views.data.couch_reqmgr_aux_db, None)])
        self.setSchemaModules([])

        super().setUp(initRoot=initRoot)

        self.setFakeDN()
        # print "%s" % self.test_authz_key.data
        # self.default_status_header = getAuthHeader(self.test_authz_key.data, PPD)

        normPath = os.path.normpath(os.path.join(os.path.dirname(__file__), '..', '..', '..', '..'))
        rerecoPath = os.path.join(normPath, 'data/ReqMgr/requests/DMWM/ReReco_RunBlockWhite.json')
        with open(rerecoPath, encoding='utf-8') as jObj:
            rerecoArgs = json.load(jObj)
        self.rerecoCreateArgs = rerecoArgs["createRequest"]
        self.rerecoCreateArgs["PrepID"] = "test_prepid"
        self.rerecoAssignArgs = rerecoArgs["assignRequest"]
        # overwrite rereco args
        self.rerecoAssignArgs["AcquisitionEra"] = "test_aqc"
        self.rerecoAssignArgs["Dashboard"] = "test"
        self.rerecoAssignArgs["ProcessingString"] = "ProcStr_ReReco"
        self.rerecoAssignArgs["SiteWhitelist"] = ["T2_CH_CERN"]

        taskChainPath = os.path.join(normPath, 'data/ReqMgr/requests/DMWM/TC_LHE_PFN.json')
        with open(taskChainPath, encoding='utf-8') as jObj:
            lheArgs = json.load(jObj)
        self.lheStep0CreateArgs = lheArgs["createRequest"]
        self.lheStep0AssignArgs = lheArgs["assignRequest"]
        self.lheStep0AssignArgs["AcquisitionEra"] = "test_aqc"

        cmsswDoc = {"_id": "software"}
        cmsswDoc[self.rerecoCreateArgs["ScramArch"]] = []
        cmsswDoc[self.rerecoCreateArgs["ScramArch"]].append(self.rerecoCreateArgs["CMSSWVersion"])
        insertDataToCouch(os.getenv("COUCHURL"), config.views.data.couch_reqmgr_aux_db, cmsswDoc)

    def tearDown(self):
        RESTBaseUnitTestWithDBBackend.tearDown(self)

    def getRequestWithNoStale(self, query):
        prefixWithNoStale = "data/request?_nostale=true&"
        return self.jsonSender.get(prefixWithNoStale + query,
                                   incoming_headers=self.ppd_header)

    def postRequestWithAuth(self, data):
        return self.jsonSender.post('data/request', data, incoming_headers=self.ppd_header)

    def putRequestWithAuth(self, requestName, data, clientHeaders):
        """
        WMCore.REST doesn take query for the put request.
        data need to send on the body
        """
        return self.jsonSender.put('data/request/%s' % requestName, data,
                                   incoming_headers=clientHeaders)

    def getMultiRequestsWithAuth(self, data):
        return self.jsonSender.post('data/request/bynames', data, incoming_headers=self.ops_header)

    def cloneRequestWithAuth(self, requestName, params=None):
        """
        WMCore.REST doesn take query for the put request.
        data need to send on the body
        """
        params = params or {}
        # params["OriginalRequestName"] = requestName
        return self.jsonSender.put('data/request/clone/%s' % requestName, params,
                                   incoming_headers=self.ops_header)

    def resultLength(self, response):
        # e.g. of response object: ({'result': []}, 200, 'OK', False)
        return len(response[0]['result'])

    def insertRequest(self, args):
        # test post method
        response = self.postRequestWithAuth(self.rerecoCreateArgs)
        self.assertEqual(response[1], 200)
        requestName = response[0]['result'][0]['request']
        return requestName

    def testRequestSimpleCycle(self):
        """
        test request cycle with one request without composite get condition.
        post, get, put
        """

        # test post method
        requestName = self.insertRequest(self.rerecoCreateArgs)

        ## test get method
        # get by name
        response = self.getRequestWithNoStale('name=%s' % requestName)
        self.assertEqual(response[1], 200, "get by name")
        self.assertEqual(self.resultLength(response), 1)

        # get by status
        response = self.getRequestWithNoStale('status=new')
        self.assertEqual(response[1], 200, "get by status")
        self.assertEqual(self.resultLength(response), 1)

        # get by prepID
        response = self.getRequestWithNoStale('prep_id=%s' % self.rerecoCreateArgs["PrepID"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 1)
        # import pdb
        # pdb.set_trace()
        response = self.getRequestWithNoStale('campaign=%s' % self.rerecoCreateArgs["Campaign"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 1)

        response = self.getRequestWithNoStale('inputdataset=%s' % self.rerecoCreateArgs["InputDataset"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 1)

        # test put request with just status change
        data = {'RequestStatus': 'assignment-approved'}
        self.putRequestWithAuth(requestName, data, self.ppd_header)
        response = self.getRequestWithNoStale('status=assignment-approved')
        self.assertEqual(response[1], 200, "put request status change")
        self.assertEqual(self.resultLength(response), 1)

        # PPD should not be able to assign requests
        data = {'RequestStatus': 'assigned'}
        data.update(self.rerecoAssignArgs)
        with self.assertRaises(HTTPException):
            self.putRequestWithAuth(requestName, data, self.ppd_header)
        response = self.getRequestWithNoStale('status=assigned')
        self.assertEqual(response[1], 200, "put request status change")
        print(response)
        self.assertEqual(self.resultLength(response), 0)

        # now try the same transition, but with Ops credentials
        self.putRequestWithAuth(requestName, data, self.ops_header)
        response = self.getRequestWithNoStale('status=assigned')
        self.assertEqual(response[1], 200, "put request status change")
        self.assertEqual(self.resultLength(response), 1)

        response = self.getRequestWithNoStale('status=assigned&team=%s' %
                                              self.rerecoAssignArgs['Team'])
        self.assertEqual(response[1], 200, "put request status change")
        self.assertEqual(self.resultLength(response), 1)

        response = self.getMultiRequestsWithAuth([requestName])
        self.assertEqual(self.resultLength(response), 1)
        self.assertEqual(list(response[0]['result'][0])[0], requestName)

    def atestRequestCombinedGetCall(self):
        """
        test request composite get call
        """
        rerecoReqName = self.insertRequest(self.rerecoCreateArgs)
        lheReqName = self.insertRequest(self.lheStep0CreateArgs)

        ## test get method and get by name
        response = self.getRequestWithNoStale('name=%s&name=%s' % (rerecoReqName, lheReqName))
        self.assertEqual(response[1], 200, "get by name")
        self.assertEqual(self.resultLength(response), 2)

        # get by status
        response = self.getRequestWithNoStale('status=new')
        self.assertEqual(response[1], 200, "get by status")
        self.assertEqual(self.resultLength(response), 2)

        # get by prepID
        response = self.getRequestWithNoStale('prep_id=%s' % self.rerecoCreateArgs["PrepID"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 2)
        # import pdb
        # pdb.set_trace()
        response = self.getRequestWithNoStale('campaign=%s' % self.rerecoCreateArgs["Campaign"])
        self.assertEqual(response[1], 200)
        self.assertEqual(self.resultLength(response), 2)

    def atestRequestClone(self):
        requestName = self.insertRequest(self.rerecoCreateArgs)
        response = self.cloneRequestWithAuth(requestName)
        print(response)
        self.assertEqual(response[1], 200, "put request clone")
        response = self.getRequestWithNoStale('status=new')
        self.assertEqual(self.resultLength(response), 2)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""
The DBSUpload algorithm

This code, all of which is in uploadDatasets works in three parts.
Each part is carried out in its own database transaction.

1) Find all files that have not yet been uploaded to DBS,
   and the fileset-algo pairs that go along with them.
2) Upload these files.  This is handled through DBSInterface
   which first sends the algo, then the dataset, and then
   commits all of the files.  It also handles migrating to
   global is you specify a global URL and the file blocks are
   full (have reached the limit in the DBS config)
3) The code cleans up at the end, check for any open blocks
   that have exceeded their timeout.


NOTE: This is complicated as hell, because you can have a
dataset-algo pair where BOTH the dataset and the algo are in
DBS attached to different objects.  You have to scan for a pair
that is not in DBS, decide whether its components are in DBS,
add them, and then add the files.  This is why everything is
so convoluted.
"""
from builtins import range

from future import standard_library
from future.utils import viewvalues

standard_library.install_aliases()

import queue
import json
import logging
import multiprocessing
import os.path
import threading
import time

from dbs.apis.dbsClient import DbsApi
from RestClient.ErrorHandling.RestClientExceptions import HTTPError

from Utils.Timers import timeFunction
from WMComponent.DBS3Buffer.DBSBufferBlock import DBSBufferBlock
from WMComponent.DBS3Buffer.DBSBufferUtil import DBSBufferUtil
from WMCore.Algorithms.MiscAlgos import sortListByKey
from WMCore.DAOFactory import DAOFactory
from WMCore.Services.UUIDLib import makeUUID
from WMCore.Services.WMStatsServer.WMStatsServer import WMStatsServer
from WMCore.Services.DBS.DBSErrors import DBSError
from WMCore.WMException import WMException
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread


def uploadWorker(workInput, results, dbsUrl, gzipEncoding=False):
    """
    _uploadWorker_

    Put JSONized blocks in the workInput
    Get confirmation in the output

    :param workInput: work input data
    :param results: output results dictionary
    :param dbsUrl: url of DBS server to use
    :param gzipEncoding: specify if we should use gzipEncoding
    """

    # Init DBS Stuff
    logging.debug("Creating dbsAPI with address %s", dbsUrl)
    dbsApi = DbsApi(url=dbsUrl, useGzip=gzipEncoding)

    while True:

        try:
            work = workInput.get()
        except (EOFError, IOError):
            crashMessage = "Hit EOF/IO in getting new work\n"
            crashMessage += "Assuming this is a graceful break attempt.\n"
            logging.error(crashMessage)
            break

        if work == 'STOP':
            # Then halt the process
            break

        name = work.get('name', None)  # this is the block name
        block = work.get('block', None)  # this is the block data structure

        # Do stuff with DBS
        try:
            logging.info("About to call insert block for: %s", name)
            dbsApi.insertBulkBlock(blockDump=block)
            results.put({'name': name, 'success': "uploaded"})
        except HTTPError as ex:
            # DBS Go server errors are defined here:
            # https://github.com/dmwm/dbs2go/blob/master/dbs/errors.go
            dbsError = DBSError(ex.body)
            reason = dbsError.getReason()
            message = dbsError.getMessage()
            srvCode = dbsError.getServerCode()
            msg = f'DBSError code: {srvCode}, message: {message}, reason: {reason}'
            if srvCode == 128:
                # block already exist
                logging.warning("Block %s already exists. Marking it as uploaded.", name)
                results.put({'name': name, 'success': "check"})
            elif srvCode in [132, 133, 134, 135, 136, 137, 138, 139, 140]:
                # racing conditions
                logging.warning("Hit a transient data race condition injecting block %s, %s", name, msg)
                results.put({'name': name, 'success': "error", 'error': msg})
            else:
                msg = f"Error trying to process block {name} through DBS. Details: {msg}"
                logging.error(msg)
                results.put({'name': name, 'success': "error", 'error': msg})
        except Exception as ex:
            msg = f"Hit a general exception while inserting block {name}. Error: {str(ex)}"
            logging.exception(msg)
            results.put({'name': name, 'success': "error", 'error': msg})
    return


def parseDBSException(exBodyString):
    """
    parse DBS Go-based server exception
    :param exBodyString: exception message body string (not exception).
    The upstream code extract HTTP body from exception object and pass it here.
    :return: either (parsed) concise exception message or original body string
    """
    try:
        data = json.loads(exBodyString)
        # dbs2go always return a list
        return data[0]['error']['reason']
    except:
        return exBodyString


class DBSUploadException(WMException):
    """
    Holds the exception info for
    all the things that will go wrong
    """


class DBSUploadPoller(BaseWorkerThread):
    """
    Handles poll-based DBSUpload

    """

    def __init__(self, config):
        """
        Initialise class members
        """
        logging.info("Running __init__ for DBS3 Uploader")
        BaseWorkerThread.__init__(self)
        self.config = config

        # This is slightly dangerous, but DBSUpload depends
        # on DBSInterface anyway
        self.dbsUrl = self.config.DBS3Upload.dbsUrl

        # Tier0 Agent don't need this
        if hasattr(self.config, "Tier0Feeder"):
            self.wmstatsServerSvc = None
        else:
            wmstatsSvcURL = self.config.General.centralWMStatsURL.replace("couchdb/wmstats",
                                                                          "wmstatsserver")
            self.wmstatsServerSvc = WMStatsServer(wmstatsSvcURL)

        self.dbsUtil = DBSBufferUtil()

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)
        self.updateBlocksDAO = daoFactory(classname="UpdateBlocks")
        self.updateFilesDAO = daoFactory(classname="UpdateFiles")
        self.createBlocksDAO = daoFactory(classname="CreateBlocks")
        self.setBlockFilesDAO = daoFactory(classname="SetBlockFiles")

        self.pool = []
        self.blocksToCheck = []
        self.workInput = None
        self.workResult = None
        self.nProc = getattr(self.config.DBS3Upload, 'nProcesses', 4)
        self.wait = getattr(self.config.DBS3Upload, 'dbsWaitTime', 2)
        self.nTries = getattr(self.config.DBS3Upload, 'dbsNTries', 300)
        self.physicsGroup = getattr(self.config.DBS3Upload, "physicsGroup", "NoGroup")
        self.datasetType = getattr(self.config.DBS3Upload, "datasetType", "PRODUCTION")
        self.primaryDatasetType = getattr(self.config.DBS3Upload, "primaryDatasetType", "mc")
        self.uploaderName = getattr(self.config.DBS3Upload, "uploaderName", "WMAgent")
        if self.uploaderName not in ("WMAgent", "T0Prod", "T0Replay"):
            raise DBSUploadException(f"Invalid value for attribute uploaderName: {self.uploaderName}")

        self.blockCount = 0
        self.gzipEncoding = getattr(self.config.DBS3Upload, 'gzipEncoding', False)
        self.dbsApi = DbsApi(url=self.dbsUrl)

        # List of blocks currently in processing
        self.queuedBlocks = []

        # Set up the pool of worker processes
        self.setupPool()

        # Setting up any cache objects
        self.blockCache = {}

        self.filesToUpdate = []

        self.dumpBlockJsonFor = getattr(self.config.DBS3Upload, 'dumpBlockJsonFor', "")

        self.copyPath = os.path.join(getattr(self.config.DBS3Upload, 'componentDir', '/data/srv/'),
                                     'dbsuploader_block.json')

        self.timeoutWaiver = 1

        self.datasetParentageCache = {}

        return

    def setupPool(self):
        """
        _setupPool_

        Set up the processing pool for work
        """
        if self.pool:
            # Then something already exists.  Continue
            return

        self.workInput = multiprocessing.Queue()
        self.workResult = multiprocessing.Queue()

        # Starting up the pool:
        for _ in range(self.nProc):
            p = multiprocessing.Process(target=uploadWorker,
                                        args=(self.workInput,
                                              self.workResult,
                                              self.dbsUrl,
                                              self.gzipEncoding))
            p.start()
            self.pool.append(p)

        return

    def __del__(self):
        """
        __del__

        Trigger a close of connections if necessary
        """
        self.close()
        return

    def close(self):
        """
        _close_

        Kill all connections and terminate
        """
        terminate = False
        for _ in self.pool:
            try:
                self.workInput.put('STOP')
            except Exception as ex:
                # Something very strange happens here
                # It's like it raises a blank exception
                # Upon being told to return
                msg = "Hit some exception in deletion\n"
                msg += str(ex)
                logging.debug(msg)
                terminate = True
        try:
            self.workInput.close()
            self.workResult.close()
        except Exception:
            # What are you going to do?
            pass
        for proc in self.pool:
            if terminate:
                proc.terminate()
            else:
                proc.join()
        self.pool = []
        self.workInput = None
        self.workResult = None
        return

    def terminate(self, parameters):
        """
        Do one more pass, then terminate

        """
        logging.debug("terminating. doing one more pass before we die")
        self.algorithm(parameters)

    @timeFunction
    def algorithm(self, parameters=None):
        """
        _algorithm_

        First, check blocks that may be already uploaded
        Then, load blocks
        Then, load files
        Then, move files into blocks
        Then add new blocks in DBSBuffer
        Then add blocks to DBS
        Then mark blocks as done in DBSBuffer
        """
        logging.info("Starting the DBSUpload Polling Cycle")
        # refreshing parentageCache every cycle
        if self.updateDatasetParentageCache() is False:
            return

        logging.debug("Dataset parentage map: %s", self.datasetParentageCache)
        try:
            self.checkBlocks()
            self.loadBlocks()
            self.loadFiles()
            self.checkBlockCompletion()
            self.inputBlocks()
            self.retrieveBlocks()
        except WMException:
            raise
        except Exception as ex:
            msg = "Unhandled Exception in DBSUploadPoller! Error: %s" % str(ex)
            logging.exception(msg)
            raise DBSUploadException(msg) from None

    def updateDatasetParentageCache(self):
        """
        Return True to indicate it successfully fetched the parentage
        map. If there was an exception, return False
        """
        myThread = threading.currentThread()

        success = True
        if not self.wmstatsServerSvc:
            self.datasetParentageCache = {}
            return success

        try:
            self.datasetParentageCache = self.wmstatsServerSvc.getChildParentDatasetMap()
        except Exception as ex:
            success = False
            excReason = getattr(ex, 'reason', '')
            errorMsg = 'Failed to fetch parentage map from WMStats, skipping this cycle. '
            errorMsg += 'Reason: {}. Error: {}. '.format(excReason, str(ex))
            logging.error(errorMsg)
            myThread.logdbClient.post("DBS3Upload_parentMap", errorMsg, "warning")
        else:
            myThread.logdbClient.delete("DBS3Upload_parentMap", "warning", this_thread=True)

        return success

    def loadBlocks(self):
        """
        _loadBlocks_

        Find all blocks; make sure they're in the cache
        """
        logging.info("Executing loadBlocks method...")
        openBlocks = self.dbsUtil.findOpenBlocks()
        logging.info("Found %d open blocks.", len(openBlocks))
        logging.debug("These are the openblocks: %s", openBlocks)

        # Load them if we don't have them
        blocksToLoad = []
        for block in openBlocks:
            if block['blockname'] not in self.blockCache:
                blocksToLoad.append(block['blockname'])

        # Now load the blocks
        try:
            logging.info("Now loading %d open blocks not yet cached...", len(blocksToLoad))
            loadedBlocks = self.dbsUtil.loadBlocks(blocksToLoad)
            logging.info("Loaded %d blocks from the database.", len(loadedBlocks))
        except WMException:
            raise
        except Exception as ex:
            msg = "Unhandled exception while loading blocks.\n"
            msg += str(ex)
            logging.error(msg)
            logging.debug("Blocks to load: %s\n", blocksToLoad)
            raise DBSUploadException(msg) from None

        for blockInfo in loadedBlocks:
            logging.info("Creating DBSBufferBlock object for: %s", blockInfo['block_name'])
            block = DBSBufferBlock(name=blockInfo['block_name'],
                                   location=blockInfo['origin_site_name'],
                                   datasetpath=blockInfo['datasetpath'],
                                   uploader=self.uploaderName)

            parent = self.datasetParentageCache.get(blockInfo['datasetpath'])
            if parent:
                block.addDatasetParent(parent)
                logging.debug("Load block: Child dataset %s, Parent dataset %s", blockInfo['datasetpath'], parent)
            block.FillFromDBSBuffer(blockInfo)
            blockname = block.getName()

            # Now we load the dataset information
            self.setDatasetInfo(block)

            # Now we have to load files...
            try:
                files = self.dbsUtil.loadFilesByBlock(blockname=blockname)
                logging.info("Have %i files for block %s", len(files), blockname)
            except WMException:
                raise
            except Exception as ex:
                msg = "Unhandled exception while loading files for existing blocks.\n"
                msg += str(ex)
                logging.error(msg)
                logging.debug("Blocks being loaded: %s\n", blockname)
                raise DBSUploadException(msg) from None

            # Add the loaded files to the block
            for f in files:
                block.addFile(f, self.datasetType, self.primaryDatasetType)

            # Add to the cache
            self.blockCache[blockInfo['block_name']] = block

        return

    def loadFiles(self):
        """
        _loadFiles_

        Load all files that need to be loaded.  I will do this by DatasetPath
        to break the monstrous calls down into smaller chunks.
        """
        logging.info("Executing loadFiles method...")
        dspList = self.dbsUtil.findUploadableDAS()

        readyBlocks = []
        for dspInfo in dspList:

            datasetpath = dspInfo['DatasetPath']

            # Get the files
            try:
                loadedFiles = self.dbsUtil.findUploadableFilesByDAS(datasetpath=datasetpath)
            except WMException:
                raise
            except Exception as ex:
                msg = "Unhandled exception while loading uploadable files for DatasetPath.\n"
                msg += str(ex)
                logging.error(msg)
                logging.debug("DatasetPath being loaded: %s\n", datasetpath)
                raise DBSUploadException(msg) from None

            # Sort the files and blocks by location
            fileDict = sortListByKey(loadedFiles, 'locations')

            # Now add each file
            for location in fileDict:

                files = fileDict.get(location)

                if files:
                    currentBlock = self.getBlock(files[0], location, True)
                    currentBlock.setAcquisitionEra(era=dspInfo['AcquisitionEra'])
                    currentBlock.setProcessingVer(procVer=dspInfo['ProcessingVer'])

                    for newFile in files:

                        if not newFile.get('block', 1) is None:
                            # Then this file already has a block
                            # It should be accounted for somewhere
                            # Or loaded with the block
                            continue

                        # Check if we can put files in this block
                        if not self.isBlockOpen(newFile=newFile,
                                                block=currentBlock):
                            # Then we have to close the block and get a new one
                            currentBlock.setPendingAndCloseBlock()
                            readyBlocks.append(currentBlock)
                            currentBlock = self.getBlock(newFile=newFile,
                                                         location=location)
                            currentBlock.setAcquisitionEra(era=dspInfo['AcquisitionEra'])
                            currentBlock.setProcessingVer(procVer=dspInfo['ProcessingVer'])

                        # Now deal with the file
                        currentBlock.addFile(newFile, self.datasetType, self.primaryDatasetType)
                        self.filesToUpdate.append({'filelfn': newFile['lfn'],
                                                   'block': currentBlock.getName()})
                    # Done with the location
                    readyBlocks.append(currentBlock)

        for block in readyBlocks:
            self.blockCache[block.getName()] = block

        return

    def checkBlockCompletion(self):
        """
        _checkBlockCompletion_

        Mark Open blocks as Pending if they have timed out or their workflows have completed
        """
        logging.info("Executing checkBlockCompletion method...")
        completedWorkflows = self.dbsUtil.getCompletedWorkflows()
        for block in viewvalues(self.blockCache):
            if block.status == "Open":
                if (block.getTime() > block.getMaxBlockTime()) or any(
                        key in completedWorkflows for key in block.workflows):
                    block.setPendingAndCloseBlock()

        return

    def isBlockOpen(self, newFile, block, doTime=False):
        """
        _isBlockOpen_

        Check and see if a block is full
        This will check on time, but that's disabled by default
        The plan is to do a time check after we do everything else,
        so open blocks about to time out can still get more
        files put in them.
        """

        if block.getMaxBlockFiles() is None or block.getMaxBlockNumEvents() is None or \
                block.getMaxBlockSize() is None or block.getMaxBlockTime() is None:
            return True
        if block.status != 'Open':
            # Then somebody has dumped this already
            return False
        if block.getSize() + newFile['size'] > block.getMaxBlockSize():
            return False
        if block.getNumEvents() + newFile['events'] > block.getMaxBlockNumEvents():
            return False
        if block.getNFiles() >= block.getMaxBlockFiles():
            # Then we have to dump it because this file
            # will put it over the limit.
            return False
        if block.getTime() > block.getMaxBlockTime() and doTime:
            return False

        return True

    def getBlock(self, newFile, location, skipOpenCheck=False):
        """
        _getBlock_

        Retrieve a block is one exists with matching datasetpath/location and is open.
        If no such block is found create and return a new one.
        """
        datasetpath = newFile["datasetPath"]

        for block in viewvalues(self.blockCache):
            if datasetpath == block.getDatasetPath() and location == block.getLocation():
                if not self.isBlockOpen(newFile=newFile, block=block) and not skipOpenCheck:
                    # Block isn't open anymore.  Mark it as pending so that it gets uploaded.
                    block.setPendingAndCloseBlock()
                else:
                    return block

        # A suitable open block does not exist.  Create a new one.
        blockname = "%s#%s" % (datasetpath, makeUUID())
        newBlock = DBSBufferBlock(name=blockname,
                                  location=location,
                                  datasetpath=datasetpath,
                                  uploader=self.uploaderName)
        # Now we load the dataset information
        self.setDatasetInfo(newBlock)

        parent = self.datasetParentageCache.get(datasetpath)
        if parent:
            newBlock.addDatasetParent(parent)
            logging.debug("Get block: Child dataset %s, Parent dataset %s", datasetpath, parent)

        self.blockCache[blockname] = newBlock
        return newBlock

    def setDatasetInfo(self, blockObj):
        """
        Given a block object, look up its dataset and set the necessary
        data structure in the object
        :param blockObj: a DBSBufferBlock object
        :return: None, the object is updated here
        """
        try:
            dsetInfo = self.dbsUtil.loadDataset(blockObj.getDatasetPath())
            blockObj.setDataset(datasetName=dsetInfo['path'],
                                primaryType=self.primaryDatasetType,
                                datasetType=dsetInfo['valid_status'],
                                prep_id=dsetInfo['prep_id'])
        except Exception as ex:
            msg = f"Unhandled exception while loading/setting dataset for block: {blockObj.getName()}. "
            msg += f"Details: {str(ex)}"
            raise DBSUploadException(msg) from None

    def inputBlocks(self):
        """
        _inputBlocks_

        Loop through all of the "active" blocks and sort them so we can act
        appropriately on them.  Everything will be sorted based on the
        following:
         Queued - Block is already being acted on by another process.  We just
          ignore it.
         Pending, not in DBSBuffer - Block that has been closed and needs to
           be injected into DBS and also written to DBSBuffer.  We'll do both.
         Pending, in DBSBuffer - Block has been closed and written to
           DBSBuffer.  We just need to inject it into DBS.
         Open, not in DBSBuffer - Newly created block that needs to be written
           not DBSBuffer.
         Open, in DBSBuffer - Newly created block that has already been
           written to DBSBuffer.  We don't have to do anything with it.
        """
        logging.info("Executing inputBlocks method...")
        if not self.blockCache:
            return

        myThread = threading.currentThread()

        createInDBS = []
        createInDBSBuffer = []
        updateInDBSBuffer = []
        for block in viewvalues(self.blockCache):
            if block.getName() in self.queuedBlocks:
                # Block is already being dealt with by another process.  We'll
                # ignore it here.
                continue
            if block.status == 'Pending':
                # All pending blocks need to be injected into DBS.
                createInDBS.append(block)

                # If this is a new block it needs to be added to DBSBuffer
                # otherwise it just needs to be updated in DBSBuffer.
                if not block.inBuff:
                    createInDBSBuffer.append(block)
                else:
                    updateInDBSBuffer.append(block)
            if block.status == 'Open' and not block.inBuff:
                # New block that needs to be added to DBSBuffer.
                createInDBSBuffer.append(block)

        # First handle new and updated blocks
        if createInDBSBuffer or updateInDBSBuffer:
            try:
                myThread.transaction.begin()
                if createInDBSBuffer:
                    logging.info("Creating %d new blocks in DBSBuffer", len(createInDBSBuffer))
                    self.createBlocksDAO.execute(blocks=createInDBSBuffer,
                                                 conn=myThread.transaction.conn,
                                                 transaction=True)
                if updateInDBSBuffer:
                    logging.info("Updating %d blocks in DBSBuffer", len(updateInDBSBuffer))
                    self.updateBlocksDAO.execute(blocks=updateInDBSBuffer,
                                                 conn=myThread.transaction.conn,
                                                 transaction=True)
            except WMException:
                myThread.transaction.rollback()
                raise
            except Exception as ex:
                myThread.transaction.rollback()
                msg = "Unhandled exception while writing new blocks into DBSBuffer\n"
                msg += str(ex)
                logging.error(msg)
                logging.debug("Blocks for DBSBuffer: %s\n", createInDBSBuffer)
                logging.debug("Blocks for Update: %s\n", updateInDBSBuffer)
                raise DBSUploadException(msg) from None
            else:
                myThread.transaction.commit()

        # Update block status in the block cache.  Mark the blocks that we have
        # added to DBSBuffer as being in DBSBuffer.
        for block in createInDBSBuffer:
            self.blockCache.get(block.getName()).inBuff = True

        # Record new file/block associations in DBSBuffer.
        if self.filesToUpdate:
            try:
                myThread.transaction.begin()
                logging.info("Associating %d files to blocks in DBSBuffer", len(self.filesToUpdate))
                self.setBlockFilesDAO.execute(binds=self.filesToUpdate,
                                              conn=myThread.transaction.conn,
                                              transaction=True)
                self.filesToUpdate = []
            except WMException:
                myThread.transaction.rollback()
                raise
            except Exception as ex:
                myThread.transaction.rollback()
                msg = "Unhandled exception while setting blocks in files.\n"
                msg += str(ex)
                logging.error(msg)
                logging.debug("Files to Update: %s\n", self.filesToUpdate)
                raise DBSUploadException(msg) from None
            else:
                myThread.transaction.commit()

        if not createInDBS:
            # then there is nothing else to do
            return

        # Build the pool if it was closed
        if not self.pool:
            self.setupPool()

        # Finally upload blocks to DBS.
        for block in createInDBS:
            if not block.files:
                # What are we doing?
                logging.debug("Skipping empty block")
                continue

            logging.debug("Found block %s in blocks", block.getName())
            block.setPhysicsGroup(group=self.physicsGroup)

            encodedBlock = block.convertToDBSBlock()
            logging.info("Queueing block for insertion: %s", block.getName())
            self.workInput.put({'name': block.getName(), 'block': encodedBlock})
            self.blockCount += 1
            if self.dumpBlockJsonFor and (self.dumpBlockJsonFor == block.getName()):
                logging.info("Dumping '%s' information into %s", block.getName(), self.copyPath)
                with open(self.copyPath, 'w') as jo:
                    json.dump(encodedBlock, jo, indent=2)
            self.queuedBlocks.append(block.getName())

        # And all work is in and we're done for now
        return

    def retrieveBlocks(self):
        """
        _retrieveBlocks_

        Once blocks are in DBS, we have to retrieve them and see what's
        in them.  What we do is get everything out of the result queue,
        and then update it in DBSBuffer.

        To do this, the result queue needs to pass back the blockname
        """
        logging.info("Executing retrieveBlocks method...")
        myThread = threading.currentThread()

        blocksToClose = []
        emptyCount = 0
        while self.blockCount > 0:
            if emptyCount > self.nTries:

                # When timeoutWaiver is 0 raise error.
                # It could take long time to get upload data to DBS
                # if there are a lot of files are cumulated in the buffer.
                # in first try but second try should be faster.
                # timeoutWaiver is set as component variable - only resets when component restarted.
                # The reason for that is only back log will occur when component is down
                # for a long time while other component still running and feeding the data to
                # dbsbuffer

                if self.timeoutWaiver == 0:
                    msg = "Exceeded max number of waits while waiting for DBS to finish"
                    raise DBSUploadException(msg)
                else:
                    self.timeoutWaiver = 0
                    return
            try:
                # Get stuff out of the queue with a ridiculously
                # short wait time
                blockresult = self.workResult.get(timeout=self.wait)
                blocksToClose.append(blockresult)
                self.blockCount -= 1
                logging.debug("Got a block to close")
            except queue.Empty:
                # This means the queue has no current results
                time.sleep(2)
                emptyCount += 1
                continue

        loadedBlocks = []
        for result in blocksToClose:
            # Remove from list of work being processed
            self.queuedBlocks.remove(result.get('name'))
            if result["success"] == "uploaded":
                block = self.blockCache.get(result.get('name'))
                block.status = 'InDBS'
                loadedBlocks.append(block)
            elif result["success"] == "check":
                block = result["name"]
                self.blocksToCheck.append(block)

        if loadedBlocks:
            try:
                logging.info("Updating database record with files injected into DBS")
                myThread.transaction.begin()
                self.updateFilesDAO.execute(blocks=loadedBlocks, status="InDBS",
                                            conn=myThread.transaction.conn,
                                            transaction=True)
                logging.info("Updating %d blocks successfully injected into DBS", len(loadedBlocks))
                self.updateBlocksDAO.execute(blocks=loadedBlocks,
                                             conn=myThread.transaction.conn,
                                             transaction=True)
            except Exception as ex:
                myThread.transaction.rollback()
                # possible deadlock with PhEDExInjector, retry once after 10s
                logging.warning("Oracle exception, possible deadlock due to race condition, retry after 10s sleep")
                time.sleep(10)
                try:
                    myThread.transaction.begin()
                    self.updateFilesDAO.execute(blocks=loadedBlocks, status="InDBS",
                                                conn=myThread.transaction.conn,
                                                transaction=True)
                    self.updateBlocksDAO.execute(blocks=loadedBlocks,
                                                 conn=myThread.transaction.conn,
                                                 transaction=True)
                except Exception as ex:
                    myThread.transaction.rollback()
                    msg = "Unhandled exception while finished closed blocks in DBSBuffer\n"
                    msg += str(ex)
                    logging.error(msg)
                    logging.debug("Blocks for Update: %s\n", loadedBlocks)
                    raise DBSUploadException(msg) from None
                else:
                    myThread.transaction.commit()

            else:
                myThread.transaction.commit()

        for block in loadedBlocks:
            # Clean things up
            name = block.getName()
            del self.blockCache[name]

        # Clean up the pool so we don't have stuff waiting around
        if self.pool:
            self.close()

        # And we're done
        return

    def checkBlocks(self):
        """
        _checkBlocks_

        Check with DBS3 if the blocks marked as check are
        uploaded or not.
        """
        logging.info("Checking for blocks potentially uploaded but not marked as such in the database.")
        myThread = threading.currentThread()

        blocksUploaded = []

        # See if there is anything to check
        for block in self.blocksToCheck:
            logging.debug("Checking block existence: %s", block)
            # Check in DBS if the block was really inserted
            try:
                # FIXME: is it still an empty list for blocks not available?
                result = self.dbsApi.listBlocks(block_name=block)
                # it is an empty list if block cannot be found
                if result:
                    loadedBlock = self.blockCache.get(block)
                    loadedBlock.status = 'InDBS'
                    blocksUploaded.append(loadedBlock)
                    logging.info("Block '%s' will be marked as '%s'", block, loadedBlock.status)
            except Exception as ex:
                msg = "Error trying to check block %s through DBS. Error: %s" % (block, str(ex))
                logging.exception(msg)

        # Update the status of those blocks that were truly inserted
        if blocksUploaded:
            try:
                logging.info("Marking bulk of blocks as 'InDBS'")
                myThread.transaction.begin()
                self.updateBlocksDAO.execute(blocks=blocksUploaded,
                                             conn=myThread.transaction.conn,
                                             transaction=True)
                self.updateFilesDAO.execute(blocks=blocksUploaded, status="InDBS",
                                            conn=myThread.transaction.conn,
                                            transaction=True)
            except WMException:
                myThread.transaction.rollback()
                raise
            except Exception as ex:
                myThread.transaction.rollback()
                msg = "Unhandled exception while finished closed blocks in DBSBuffer\n"
                msg += str(ex)
                logging.exception(msg)
                logging.debug("Blocks for Update: %s\n", blocksUploaded)
                raise DBSUploadException(msg) from None
            else:
                myThread.transaction.commit()
                logging.info("A total of %d blocks have been successfully updated to 'InDBS'",
                             len(blocksUploaded))

        for block in blocksUploaded:
            # Clean things up
            name = block.getName()
            del self.blockCache[name]

        # Clean the check list
        self.blocksToCheck = []

        # We're done
        return



#!/bin/env python
# pylint: disable=E1101, W0201, E1103
# E1101: reference config file variables
# W0201: Don't much around with __init__
# E1103: Use thread members

from __future__ import print_function
from builtins import range, object

import cProfile
import os
import pickle
import pstats
import random
import threading
import time
import unittest

from Utils.PythonVersion import PY3

from WMCore_t.WMSpec_t.TestSpec import createTestWorkload
from nose.plugins.attrib import attr

from WMComponent.JobCreator.JobCreatorPoller import JobCreatorPoller, capResourceEstimates
from WMCore.Agent.HeartbeatAPI import HeartbeatAPI
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.ResourceControl.ResourceControl import ResourceControl
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMSpec.Makers.TaskMaker import TaskMaker
from WMQuality.Emulators import EmulatorSetup
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit


class JobCreatorTest(EmulatedUnitTestCase):
    """
    Test case for the JobCreator

    """

    sites = ['T2_US_Florida', 'T2_US_UCSD', 'T2_TW_Taiwan', 'T1_CH_CERN']

    def setUp(self):
        """
        _setUp_

        Setup the database and logging connection.  Try to create all of the
        WMBS tables.  Also, create some dummy locations.
        """
        super(JobCreatorTest, self).setUp()

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()

        self.testInit.setSchema(customModules=['WMCore.WMBS', 'WMCore.ResourceControl', 'WMCore.Agent.Database'],
                                useDefault=False)
        self.couchdbname = "jobcreator_t"
        self.testInit.setupCouch("%s/jobs" % self.couchdbname, "JobDump")
        self.testInit.setupCouch("%s/fwjrs" % self.couchdbname, "FWJRDump")
        self.configFile = EmulatorSetup.setupWMAgentConfig()

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)

        locationAction = self.daoFactory(classname="Locations.New")
        for site in self.sites:
            locationAction.execute(siteName=site, pnn=site)

        # Create sites in resourceControl

        resourceControl = ResourceControl()
        for site in self.sites:
            resourceControl.insertSite(siteName=site, pnn=site, ceName=site)
            resourceControl.insertThreshold(siteName=site, taskType='Processing', maxSlots=10000, pendingSlots=10000)

        self.resourceControl = resourceControl

        self._setup = True
        self._teardown = False

        self.testDir = self.testInit.generateWorkDir()
        self.cwd = os.getcwd()

        # Set heartbeat
        self.componentName = 'JobCreator'
        self.heartbeatAPI = HeartbeatAPI(self.componentName)
        self.heartbeatAPI.registerComponent()

        if PY3:
            self.assertItemsEqual = self.assertCountEqual

        return

    def tearDown(self):
        """
        _tearDown_

        Drop all the WMBS tables.
        """

        self.testInit.clearDatabase(modules=['WMCore.WMBS', 'WMCore.ResourceControl', 'WMCore.Agent.Database'])

        self.testInit.delWorkDir()

        self._teardown = True

        self.testInit.tearDownCouch()
        EmulatorSetup.deleteConfig(self.configFile)

        return

    def createJobCollection(self, name, nSubs, nFiles, workflowURL='test'):
        """
        _createJobCollection_

        Create a collection of jobs
        """

        myThread = threading.currentThread()

        testWorkflow = Workflow(spec=workflowURL, owner="mnorman",
                                name=name, task="/TestWorkload/ReReco")
        testWorkflow.create()

        for sub in range(nSubs):

            nameStr = '%s-%i' % (name, sub)

            myThread.transaction.begin()

            testFileset = Fileset(name=nameStr)
            testFileset.create()

            for f in range(nFiles):
                # pick a random site
                site = random.choice(self.sites)
                testFile = File(lfn="/lfn/%s/%i" % (nameStr, f), size=1024, events=10)
                testFile.setLocation(site)
                testFile.create()
                testFileset.addFile(testFile)

            testFileset.commit()
            testSubscription = Subscription(fileset=testFileset,
                                            workflow=testWorkflow,
                                            type="Processing",
                                            split_algo="FileBased")
            testSubscription.create()

            myThread.transaction.commit()

        return

    def createWorkload(self, workloadName='Test'):
        """
        Creates a test workload for us to run on, hold the basic necessities.
        """

        workload = createTestWorkload(workloadName)
        rereco = workload.getTask("ReReco")
        seederDict = {"generator.initialSeed": 1001, "evtgenproducer.initialSeed": 1001}
        rereco.addGenerator("PresetSeeder", **seederDict)

        taskMaker = TaskMaker(workload, os.path.join(self.testDir, 'workloadTest'))
        taskMaker.skipSubscription = True
        taskMaker.processWorkload()

        return workload

    def getConfig(self):
        """
        _getConfig_

        Creates a common config.
        """

        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        # First the general stuff
        config.section_("General")
        config.General.workDir = os.getenv("TESTDIR", os.getcwd())

        config.section_("Agent")
        config.Agent.componentName = self.componentName

        # Now the CoreDatabase information
        # This should be the dialect, dburl, etc
        config.section_("CoreDatabase")
        config.CoreDatabase.connectUrl = os.getenv("DATABASE")
        config.CoreDatabase.socket = os.getenv("DBSOCK")

        config.component_("JobCreator")
        config.JobCreator.namespace = 'WMComponent.JobCreator.JobCreator'
        # The log level of the component.
        # config.JobCreator.logLevel = 'SQLDEBUG'
        config.JobCreator.logLevel = 'INFO'

        # maximum number of threads we want to deal
        # with messages per pool.
        config.JobCreator.maxThreads = 1
        config.JobCreator.UpdateFromResourceControl = True
        config.JobCreator.pollInterval = 10
        # config.JobCreator.jobCacheDir               = self.testDir
        config.JobCreator.defaultJobType = 'processing'  # Type of jobs that we run, used for resource control
        config.JobCreator.workerThreads = 4
        config.JobCreator.componentDir = self.testDir
        config.JobCreator.useWorkQueue = True
        config.JobCreator.WorkQueueParams = {'emulateDBSReader': True}

        # We now call the JobMaker from here
        config.component_('JobMaker')
        config.JobMaker.logLevel = 'INFO'
        config.JobMaker.namespace = 'WMCore.WMSpec.Makers.JobMaker'
        config.JobMaker.maxThreads = 1
        config.JobMaker.makeJobsHandler = 'WMCore.WMSpec.Makers.Handlers.MakeJobs'

        # JobStateMachine
        config.component_('JobStateMachine')
        config.JobStateMachine.couchurl = os.getenv('COUCHURL', 'cmssrv52.fnal.gov:5984')
        config.JobStateMachine.couchDBName = self.couchdbname

        return config

    def testVerySimpleTest(self):
        """
        _VerySimpleTest_

        Just test that everything works...more or less
        """

        # return

        myThread = threading.currentThread()

        config = self.getConfig()

        name = makeUUID()
        nSubs = 5
        nFiles = 10
        workloadName = 'TestWorkload'

        dummyWorkload = self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', 'TestWorkload', 'WMSandbox', 'WMWorkload.pkl')

        self.createJobCollection(name=name, nSubs=nSubs, nFiles=nFiles, workflowURL=workloadPath)

        testJobCreator = JobCreatorPoller(config=config)

        # First, can we run once without everything crashing?
        testJobCreator.algorithm()

        getJobsAction = self.daoFactory(classname="Jobs.GetAllJobs")
        result = getJobsAction.execute(state='Created', jobType="Processing")

        self.assertEqual(len(result), nSubs * nFiles)

        # Count database objects
        result = myThread.dbi.processData('SELECT * FROM wmbs_sub_files_acquired')[0].fetchall()
        self.assertEqual(len(result), nSubs * nFiles)

        # Find the test directory
        testDirectory = os.path.join(self.testDir, 'jobCacheDir', 'TestWorkload', 'ReReco')
        # It should have at least one jobGroup
        self.assertTrue('JobCollection_1_0' in os.listdir(testDirectory))
        # But no more then twenty
        self.assertTrue(len(os.listdir(testDirectory)) <= 20)

        groupDirectory = os.path.join(testDirectory, 'JobCollection_1_0')

        # First job should be in here
        listOfDirs = []
        for tmpDirectory in os.listdir(testDirectory):
            listOfDirs.extend(os.listdir(os.path.join(testDirectory, tmpDirectory)))
        self.assertTrue('job_1' in listOfDirs)
        self.assertTrue('job_2' in listOfDirs)
        self.assertTrue('job_3' in listOfDirs)
        jobDir = os.listdir(groupDirectory)[0]
        jobFile = os.path.join(groupDirectory, jobDir, 'job.pkl')
        self.assertTrue(os.path.isfile(jobFile))
        f = open(jobFile, 'rb')
        job = pickle.load(f)
        f.close()

        self.assertEqual(job.baggage.PresetSeeder.generator.initialSeed, 1001)
        self.assertEqual(job.baggage.PresetSeeder.evtgenproducer.initialSeed, 1001)

        self.assertEqual(job['workflow'], name)
        self.assertEqual(len(job['input_files']), 1)
        self.assertEqual(os.path.basename(job['sandbox']), 'TestWorkload-Sandbox.tar.bz2')

        return

    def testCampaignName(self):
        """
        Test campaign name is written into job pickle file
        """
        myThread = threading.currentThread()

        config = self.getConfig()

        name = makeUUID()
        nSubs = 1
        nFiles = 1
        workloadName = 'TestWorkload'

        dummyWorkload = self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', 'TestWorkload', 'WMSandbox', 'WMWorkload.pkl')

        self.createJobCollection(name=name, nSubs=nSubs, nFiles=nFiles, workflowURL=workloadPath)

        testJobCreator = JobCreatorPoller(config=config)

        # First, can we run once without everything crashing?
        testJobCreator.algorithm()

        getJobsAction = self.daoFactory(classname="Jobs.GetAllJobs")
        result = getJobsAction.execute(state='Created', jobType="Processing")

        # Find the test directory
        testDirectory = os.path.join(self.testDir, 'jobCacheDir', 'TestWorkload', 'ReReco')
        groupDirectory = os.path.join(testDirectory, 'JobCollection_1_0')

        # Get job pickle file
        jobDir = os.listdir(groupDirectory)[0]
        jobFile = os.path.join(groupDirectory, jobDir, 'job.pkl')
        self.assertTrue(os.path.isfile(jobFile))
        with open(jobFile, 'rb') as f:
            job = pickle.load(f)

        # Attribute campaign name should exist
        # but be set to the default value: None
        self.assertEqual(job['campaignName'], None)

        return

    def testPhysicsType(self):
        """
        Test physics type is written into job pickle file
        """
        myThread = threading.currentThread()

        config = self.getConfig()

        name = makeUUID()
        nSubs = 1
        nFiles = 1
        workloadName = 'TestWorkload'

        dummyWorkload = self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', 'TestWorkload', 'WMSandbox', 'WMWorkload.pkl')

        self.createJobCollection(name=name, nSubs=nSubs, nFiles=nFiles, workflowURL=workloadPath)

        testJobCreator = JobCreatorPoller(config=config)

        # First, can we run once without everything crashing?
        testJobCreator.algorithm()

        getJobsAction = self.daoFactory(classname="Jobs.GetAllJobs")
        result = getJobsAction.execute(state='Created', jobType="Processing")

        # Find the test directory
        testDirectory = os.path.join(self.testDir, 'jobCacheDir', 'TestWorkload', 'ReReco')
        groupDirectory = os.path.join(testDirectory, 'JobCollection_1_0')

        # Get job pickle file
        jobDir = os.listdir(groupDirectory)[0]
        jobFile = os.path.join(groupDirectory, jobDir, 'job.pkl')
        self.assertTrue(os.path.isfile(jobFile))
        with open(jobFile, 'rb') as f:
            job = pickle.load(f)

        # Attribute campaign name should exist
        # but be set to the default value: None
        self.assertEqual(job['physicsTaskType'], None)

        return
    @attr('performance', 'integration')
    def testProfilePoller(self):
        """
        Profile your performance
        You shouldn't be running this normally because it doesn't do anything
        """

        name = makeUUID()
        nSubs = 5
        nFiles = 1500
        workloadName = 'TestWorkload'

        workload = self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', 'TestWorkload', 'WMSandbox', 'WMWorkload.pkl')

        self.createJobCollection(name=name, nSubs=nSubs, nFiles=nFiles, workflowURL=workloadPath)

        config = self.getConfig()

        testJobCreator = JobCreatorPoller(config=config)
        cProfile.runctx("testJobCreator.algorithm()", globals(), locals(), filename="testStats.stat")

        getJobsAction = self.daoFactory(classname="Jobs.GetAllJobs")
        result = getJobsAction.execute(state='Created', jobType="Processing")

        time.sleep(10)

        self.assertEqual(len(result), nSubs * nFiles)

        p = pstats.Stats('testStats.stat')
        p.sort_stats('cumulative')
        p.print_stats(.2)

        return

    @attr('integration')
    def testProfileWorker(self):
        """
        Profile where the work actually gets done
        You shouldn't be running this one either, since it doesn't test anything.
        """

        name = makeUUID()
        nSubs = 5
        nFiles = 500
        workloadName = 'TestWorkload'

        workload = self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', 'TestWorkload', 'WMSandbox', 'WMWorkload.pkl')

        self.createJobCollection(name=name, nSubs=nSubs, nFiles=nFiles, workflowURL=workloadPath)

        config = self.getConfig()

        configDict = {"couchURL": config.JobStateMachine.couchurl,
                      "couchDBName": config.JobStateMachine.couchDBName,
                      'jobCacheDir': config.JobCreator.jobCacheDir,
                      'defaultJobType': config.JobCreator.defaultJobType}

        subs = [{"subscription": 1}, {"subscription": 2}, {"subscription": 3}, {"subscription": 4},
                {"subscription": 5}]

        testJobCreator = JobCreatorPoller(**configDict)
        cProfile.runctx("testJobCreator.algorithm(parameters = input)", globals(), locals(), filename="workStats.stat")

        p = pstats.Stats('workStats.stat')
        p.sort_stats('cumulative')
        p.print_stats(.2)

        return

    @attr('integration')
    def testHugeTest(self):
        """
        Don't run this one either

        """

        myThread = threading.currentThread()

        config = self.getConfig()

        name = makeUUID()
        nSubs = 10
        nFiles = 5000
        workloadName = 'Tier1ReReco'

        dummyWorkload = self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', 'TestWorkload', 'WMSandbox', 'WMWorkload.pkl')

        self.createJobCollection(name=name, nSubs=nSubs, nFiles=nFiles, workflowURL=workloadPath)

        testJobCreator = JobCreatorPoller(config=config)

        # First, can we run once without everything crashing?
        startTime = time.time()
        testJobCreator.algorithm()
        stopTime = time.time()

        getJobsAction = self.daoFactory(classname="Jobs.GetAllJobs")
        result = getJobsAction.execute(state='Created', jobType="Processing")

        self.assertEqual(len(result), nSubs * nFiles)

        print("Job took %f seconds to run" % (stopTime - startTime))

        # Count database objects
        result = myThread.dbi.processData('SELECT * FROM wmbs_sub_files_acquired')[0].fetchall()
        self.assertEqual(len(result), nSubs * nFiles)

        return

    def stuffWMBS(self, workflowURL, name):
        """
        _stuffWMBS_

        Insert some dummy jobs, jobgroups, filesets, files and subscriptions
        into WMBS to test job creation.  Three completed job groups each
        containing several files are injected.  Another incomplete job group is
        also injected.  Also files are added to the "Mergeable" subscription as
        well as to the output fileset for their jobgroups.
        """
        locationAction = self.daoFactory(classname="Locations.New")
        locationAction.execute(siteName="s1", pnn="somese.cern.ch")

        mergeFileset = Fileset(name="mergeFileset")
        mergeFileset.create()
        bogusFileset = Fileset(name="bogusFileset")
        bogusFileset.create()

        mergeWorkflow = Workflow(spec=workflowURL, owner="mnorman",
                                 name=name, task="/TestWorkload/ReReco")
        mergeWorkflow.create()

        mergeSubscription = Subscription(fileset=mergeFileset,
                                         workflow=mergeWorkflow,
                                         split_algo="ParentlessMergeBySize")
        mergeSubscription.create()
        dummySubscription = Subscription(fileset=bogusFileset,
                                         workflow=mergeWorkflow,
                                         split_algo="ParentlessMergeBySize")

        file1 = File(lfn="file1", size=1024, events=1024, first_event=0,
                     locations={"somese.cern.ch"})
        file1.addRun(Run(1, *[45]))
        file1.create()
        file2 = File(lfn="file2", size=1024, events=1024, first_event=1024, locations={"somese.cern.ch"})
        file2.addRun(Run(1, *[45]))
        file2.create()
        file3 = File(lfn="file3", size=1024, events=1024, first_event=2048, locations={"somese.cern.ch"})
        file3.addRun(Run(1, *[45]))
        file3.create()
        file4 = File(lfn="file4", size=1024, events=1024, first_event=3072, locations={"somese.cern.ch"})
        file4.addRun(Run(1, *[45]))
        file4.create()

        fileA = File(lfn="fileA", size=1024, events=1024, first_event=0, locations={"somese.cern.ch"})
        fileA.addRun(Run(1, *[46]))
        fileA.create()
        fileB = File(lfn="fileB", size=1024, events=1024, first_event=1024, locations={"somese.cern.ch"})
        fileB.addRun(Run(1, *[46]))
        fileB.create()
        fileC = File(lfn="fileC", size=1024, events=1024, first_event=2048, locations={"somese.cern.ch"})
        fileC.addRun(Run(1, *[46]))
        fileC.create()

        fileI = File(lfn="fileI", size=1024, events=1024, first_event=0, locations={"somese.cern.ch"})
        fileI.addRun(Run(2, *[46]))
        fileI.create()
        fileII = File(lfn="fileII", size=1024, events=1024, first_event=1024, locations={"somese.cern.ch"})
        fileII.addRun(Run(2, *[46]))
        fileII.create()
        fileIII = File(lfn="fileIII", size=1024, events=1024, first_event=2048, locations={"somese.cern.ch"})
        fileIII.addRun(Run(2, *[46]))
        fileIII.create()
        fileIV = File(lfn="fileIV", size=1024 * 1000000, events=1024, first_event=3072, locations={"somese.cern.ch"})
        fileIV.addRun(Run(2, *[46]))
        fileIV.create()

        for fileObj in [file1, file2, file3, file4, fileA, fileB, fileC, fileI, fileII, fileIII, fileIV]:
            mergeFileset.addFile(fileObj)
            bogusFileset.addFile(fileObj)

        mergeFileset.commit()
        bogusFileset.commit()

        return

    def testTestNonProxySplitting(self):
        """
        _TestNonProxySplitting_

        Test and see if we can split things without a proxy.
        """

        config = self.getConfig()
        config.JobCreator.workerThreads = 1

        name = makeUUID()
        workloadName = 'TestWorkload'

        workload = self.createWorkload(workloadName=workloadName)

        workloadPath = os.path.join(self.testDir, 'workloadTest', 'TestWorkload', 'WMSandbox', 'WMWorkload.pkl')

        self.stuffWMBS(workflowURL=workloadPath, name=name)

        testJobCreator = JobCreatorPoller(config=config)

        testJobCreator.algorithm()

        getJobsAction = self.daoFactory(classname="Jobs.GetAllJobs")
        result = getJobsAction.execute(state='Created', jobType="Processing")

        self.assertEqual(len(result), 1)

        result = getJobsAction.execute(state='Created', jobType="Merge")
        self.assertEqual(len(result), 0)

        return

    def testCapResourceEstimates(self):
        """
        _testCapResourceEstimates_

        Test capResourceEstimates function to make sure the glideinwms
        constraints are being properly considered.
        """

        class JobGroup(object):
            """Dummy object holding a jobs attr full of jobs"""

            def __init__(self):
                self.jobs = []

        constraints = {'MaxRequestDiskKB': 20971520, 'MinRequestDiskKB': 1048576,
                       'MaxWallTimeSecs': 162000, 'MinWallTimeSecs': 3600}

        jobGroups = []
        jobGroup = JobGroup()
        jobGroup.jobs.append({'estimatedJobTime': None, 'estimatedDiskUsage': None})
        jobGroup.jobs.append({'estimatedJobTime': 0, 'estimatedDiskUsage': 0})
        jobGroup.jobs.append({'estimatedJobTime': 10000, 'estimatedDiskUsage': 10 * 1000 * 1000})
        jobGroup.jobs.append({'estimatedJobTime': 200000, 'estimatedDiskUsage': 100 * 1000 * 1000})
        jobGroups.append(jobGroup)

        capResourceEstimates(jobGroups, constraints)

        self.assertItemsEqual(jobGroup.jobs[0], {'estimatedJobTime': 3600, 'estimatedDiskUsage': 1048576})
        self.assertItemsEqual(jobGroup.jobs[1], {'estimatedJobTime': 3600, 'estimatedDiskUsage': 1048576})
        self.assertItemsEqual(jobGroup.jobs[2], {'estimatedJobTime': 10000, 'estimatedDiskUsage': 10 * 1000 * 1000})
        self.assertItemsEqual(jobGroup.jobs[3], {'estimatedJobTime': 162000, 'estimatedDiskUsage': 20971520})

        return


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""
_Queries_

This module implements the Oracle backend for the persistent threadpool.

"""



import threading

from WMCore.Database.DBFormatter import DBFormatter

from WMCore.ThreadPool.MySQL.Queries import Queries as MySQLQueries

class Queries(MySQLQueries):
    """
    _Queries_

    This module implements the Oracle backend for the persistent threadpool.

    """


    def selectWork(self, args, pooltable = 'tp_threadpool'):
        """
        Select work that is not yet being processed.
        """
        # this query takes place in a locked section so
        # we do not have to worry about multiple slaves
        # getting the same work.
        result = ''
        if pooltable in ['tp_threadpool', 'tp_threadpool_buffer_in', \
            'tp_threadpool_buffer_out']:
            sqlStr = """
SELECT min(id) FROM %s WHERE component = :component AND
thread_pool_id = :thread_pool_id AND state='queued'
        """ % (pooltable)
            result = self.execute(sqlStr, args)
        else:
            sqlStr = """
SELECT min(id) FROM %s WHERE state='queued'
        """ % (pooltable)
            result = self.execute(sqlStr, {})

        return self.formatOne(result)



    def updateWorkStatus(self, args, pooltable = 'tp_threadpool'):
        """
        Updates work status of work being processed.
        """

        # differentiate between onequeu and multi queue
        if pooltable in ['tp_threadpool', 'tp_threadpool_buffer_in', \
            'tp_threadpool_buffer_out']:
            sqlStr = """
UPDATE %s SET state='queued' WHERE component = :componentName
AND thread_pool_id = :thread_pool_id
        """ % (pooltable)
            self.execute(sqlStr, args)
        else:
            sqlStr = """
UPDATE %s SET state="queued"
        """ % (pooltable)
            self.execute(sqlStr, {})

        return


    def moveWorkFromBufferIn(self, source, target):
        """
        Moves work from buffer in to main queue or buffer out
        """
        sqlUpd1 = ''
        sqlStr1 = ''
        sqlStr2 = ''
        if source == 'tp_threadpool_buffer_in':
            sqlUpd1 = """
CURSOR %s_c1 IS
  SELECT event, component, payload, thread_pool_id FROM %s
  FOR UPDATE
            """ %(target, source)
            sqlStr1 = """
INSERT INTO %s(event,component,payload,thread_pool_id) SELECT event,component,payload,thread_pool_id FROM %s
            """ % (target, source)
            sqlStr2 = """ DELETE FROM %s """ % (source)
            #self.execute(sqlUpd1, {})
        else:
            sqlStr1 = """
INSERT INTO %s(event,payload) SELECT event,payload FROM %s
            """ % (target, source)
            sqlStr2 = """ DELETE FROM %s """ % (source)

        self.execute(sqlStr1, {})
        self.execute(sqlStr2, {})

        return

    def moveWorkToBufferOut(self, args, source, target, limit):
        """
        Moves work from buffer in or main queue to the buffer out
        table.
        """

        sqlStr1 = ''
        sqlStr2 = ''
        if source in ['tp_threadpool', 'tp_threadpool_buffer_in', \
            'tp_threadpool_buffer_out']:
            # we need a for update in the select to prevent (harmless) deadlock
            # situations with innodb
            sqlStr1 = """
INSERT INTO %s (event, component, payload, thread_pool_id)
  SELECT event, component, payload, thread_pool_id FROM
   (SELECT event, component, payload, thread_pool_id, ROWNUM rn FROM %s
   WHERE component=:component AND thread_pool_id=:thread_pool_id)
  WHERE rn < %s
   """ % (target, source, limit)
            sqlStr2 = """
DELETE FROM %s WHERE id IN
(SELECT id FROM
(SELECT id, ROWNUM rn FROM %s
 WHERE component = :component
 AND thread_pool_id = :thread_pool_id
)
 WHERE rn < %s
)
""" % (source, source, limit)
            self.execute(sqlStr1, args)
            self.execute(sqlStr2, args)
        else:
            sqlStr1 = """
INSERT INTO %s(event,payload) VALUES
 (SELECT event, payload FROM
  (SELECT event, payload, ROWNUM rn FROM %s)
 WHERE rn < %s
 )
            """ % (target, source, limit)
            sqlStr2 = """
DELETE FROM %s WHERE id IN
 (SELECT id FROM
  (
   SELECT id, ROWNUM rn FROM %s
   ORDER BY id
  )
  WHERE rn < %s
 )
            """ % (source, source, limit )
            self.execute(sqlStr1, {})
            self.execute(sqlStr2, {})

        return



    def insertThreadPoolTables(self, threadpool):
        """
        __insertThreadPoolTable

        Inserts tables for threadpool (one for each threadpool)
        when multi queue is enabled.
        """
        sqlStr = """
CREATE TABLE %s(
   id                      NUMBER(11)    NOT NULL ENABLE,
   event                   varchar2(255) NOT NULL ENABLE,
   payload                 long          NOT NULL ENABLE,
   state                   varchar2(20)  NOT NULL ENABLE,

   CONSTRAINT  %s_pk     PRIMARY KEY (id),
   CONSTRAINT  %s_state  CHECK(state IN ('queued', 'process'))
   ); """ % (threadpool, threadpool, threadpool)

        self.execute(sqlStr, {})


        seqStr = """CREATE SEQUENCE %s_seq
        start with 1
        increment by 1
        nomaxvalue
        """ %(threadpool)

        self.execute(seqStr, {})

        trgStr = """CREATE TRIGGER %s_trg
BEFORE INSERT ON %s
FOR EACH ROW
     BEGIN
        SELECT %s_seq.nextval INTO :new.id FROM dual;
     END;
        """ %(threadpool)

        self.execute(trgStr, {})

        return


    def insertWork(self, args, pooltable = 'tp_threadpool'):
        """
        Inserts work into the database in case no thread can be found.
        """

        # differentiate between onequeu and multi queue
        if pooltable in ['tp_threadpool', 'tp_threadpool_buffer_in', \
            'tp_threadpool_buffer_out']:
            sqlStr = """
INSERT INTO %s(event,component,payload,thread_pool_id)
VALUES(:event,:component,:payload,:thread_pool_id)
        """  % (pooltable)
            self.execute(sqlStr, args)
        else:
            sqlStr = """
INSERT INTO %s(event,payload)
VALUES(:event,:payload)
        """  % (pooltable)
            self.execute(sqlStr, {'event':args['event'], \
                'payload':args['payload']})

        return



#!/usr/bin/env python
"""
_URLFetcher_

Fetch urls based on the sandbox information in a WMStep

"""
import os
import re
import logging

import WMCore.WMSpec.WMStep as WMStep

from WMCore.WMSpec.Steps.Fetchers.FetcherInterface import FetcherInterface
from WMCore.Services                               import Requests

class URLFetcher(FetcherInterface):
    """
    _URLFetcher_

    Fetch files via URL access based on contents of step sandbox attribute

    """


    def __call__(self, wmTask):
        """
        Trip through steps, pull in files specified in the sandbox section
        PSet Tweaks etc

        """
        for t in wmTask.steps().nodeIterator():
            t = WMStep.WMStepHelper(t)
            stepPath = "%s/%s" % (self.workingDirectory(), t.name())
            for fileInfo in t.data.sandbox:
                # fileInfo.src is the source file
                # fileInfo.injob is where we stuck it
                match = re.search("^.*/(.*?)$", fileInfo.src)
                if (match):
                    fileSuffix = match.group(1)
                else:
                    fileSuffix = "sandboxFile.dat"

                fileTarget = "%s/%s" % (stepPath, fileSuffix)

                # Now build a Request object, make a request, and write
                # the output to a file:
                try:
                    request = Requests.Requests(fileInfo.src)
                    content = request.get('')[0]
                    with open(fileTarget, 'w') as f:
                        f.write(content)
                except IOError as ex:
                    msg =  "Could not write to fileTarget %s\n" % fileTarget
                    msg += str(ex)
                    logging.error(msg)
                    logging.debug("FileInfo: %s" % fileInfo)
                    raise
                fileInfo.injob = fileTarget

        return



"""
Hello world example using WMCore.REST handling framework.
Info class giving information about ReqMgr database.
Teams, Groups, Software versions handling for ReqMgr.

"""

import logging
import cherrypy

from WMCore.REST.Server import RESTEntity, restcall, rows
from WMCore.REST.Tools import tools
from WMCore.Services.WMStats.WMStatsReader import WMStatsReader
from WMCore.REST.Format import JSONFormat
from WMCore.REST.Error import MethodWithoutQueryString


class WMStatsInfo(RESTEntity):
    """
    This class need to move under WMStats server when wmstats server created
    """
    def __init__(self, app, api, config, mount):
        # main CouchDB database where requests/workloads are stored
        RESTEntity.__init__(self, app, api, config, mount)
        wmstats_url = "%s/%s" % (self.config.couch_host, self.config.couch_wmstats_db)
        reqdb_url = "%s/%s" % (self.config.couch_host, self.config.couch_reqmgr_db)
        self.wmstats = WMStatsReader(wmstats_url, reqdbURL=reqdb_url, reqdbCouchApp="ReqMgr")

    def validate(self, apiobj, method, api, param, safe):
        args_length = len(param.args)
        if args_length == 1:
            safe.args.append(param.args[0])
            param.args.pop()
        else:
            raise MethodWithoutQueryString
        return


    @restcall(formats = [('application/json', JSONFormat())])
    @tools.expires(secs=-1)
    def get(self, request_name):
        result = self.wmstats.getRequestSummaryWithJobInfo(request_name)
        return rows([result])






#!/usr/bin/env python
"""
Unittests for the Services/MSPileup/MSPileupUtils module
"""
import logging
import unittest

from nose.plugins.attrib import attr

from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMCore.MicroService.MSPileup.DataStructs.MSPileupObj import schema
from WMCore.Services.MSPileup.MSPileupUtils import getPileupDocs


class MSPileupUtilsTest(EmulatedUnitTestCase):
    """
    Test class for the RucioUtils module
    """

    def setUp(self):
        """
        Setup for unit tests
        """
        self.logger = logging.getLogger()
        super().setUp()

    def tearDown(self):
        """
        Nothing to be done for this case
        """

    @attr('integration')
    def testGetPileupDocs(self):
        """
        Test getting pileup docs from testbed.
        """
        pdict = schema()

        dataItem = '/MinBias_TuneCP5_14TeV-pythia8/PhaseIITDRSpring19GS-106X_upgrade2023_realistic_v2_ext1-v1/GEN-SIM'
        msPileupUrl = 'https://cmsweb-testbed.cern.ch/ms-pileup/data/pileup'

        # test without filter
        queryDict = {'query': {'pileupName': dataItem}}
        result = getPileupDocs(msPileupUrl, queryDict, method='POST')

        self.logger.info('GetPileupDocs Result: %s', result)

        self.assertGreater(len(result), 0)

        # make sure all keys are in response
        for k in pdict:
            self.assertIn(k, result[0])

        # test with a filter
        filterKeys = ['currentRSEs', 'pileupName', 'containerFraction', 'ruleIds']
        queryDict = {'query': {'pileupName': dataItem},
                     'filters': filterKeys}
        result = getPileupDocs(msPileupUrl, queryDict, method='POST')

        self.assertGreater(len(result), 0)

        self.logger.info('GetPileupDocs Result: %s', result)

        # make sure only filterKeys keys are in response
        for k in pdict:
            if k in filterKeys:
                self.assertIn(k, result[0])
            else:
                self.assertNotIn(k, result[0])


if __name__ == '__main__':
    unittest.main()
